from __future__ import annotations

import pytest

from instawow._sources.tukui import TukuiResolver
from instawow.common import Defn, Flavour
from instawow.manager_ctx import ManagerCtx
from instawow.pkg_models import Pkg


@pytest.fixture
def tukui_resolver(
    iw_manager_ctx: ManagerCtx,
):
    return TukuiResolver(iw_manager_ctx)


@pytest.mark.parametrize(
    'iw_config_values',
    Flavour,
    indirect=True,
)
@pytest.mark.parametrize('alias', ['tukui', 'elvui'])
async def test_resolve_addon(
    tukui_resolver: TukuiResolver,
    alias: str,
):
    defn = Defn('tukui', alias)

    result = await tukui_resolver.resolve_one(defn, None)

    assert type(result) is Pkg
    assert result.slug == alias


async def test_changelog_url_format(
    tukui_resolver: TukuiResolver,
):
    defn = Defn('tukui', 'tukui')

    result = await tukui_resolver.resolve_one(defn, None)

    assert result.changelog_url == 'https://api.tukui.org/v1/changelog/tukui#20.38'



from __future__ import annotations

from collections.abc import Callable, Iterator, Set
from datetime import datetime
from typing import Literal

from .. import manager_ctx, pkg_db
from ..utils import bucketise, normalise_names
from . import cataloguer

_normalise_search_terms = normalise_names('')


async def search(
    manager_ctx: manager_ctx.ManagerCtx,
    search_terms: str,
    *,
    limit: int,
    sources: Set[str] = frozenset(),
    prefer_source: str | None = None,
    start_date: datetime | None = None,
    filter_installed: Literal[
        'ident', 'include_only', 'exclude', 'exclude_from_all_sources'
    ] = 'ident',
) -> list[cataloguer.ComputedCatalogueEntry]:
    "Search the catalogue for packages by name."
    import rapidfuzz
    import sqlalchemy as sa

    catalogue = await manager_ctx.synchronise()

    ew = 0.5
    dw = 1 - ew

    threshold = 0 if search_terms == '*' else 70

    if sources:
        unknown_sources = sources - manager_ctx.resolvers.keys()
        if unknown_sources:
            raise ValueError(f'Unknown sources: {", ".join(unknown_sources)}')

    if prefer_source and prefer_source not in manager_ctx.resolvers:
        raise ValueError(f'Unknown preferred source: {prefer_source}')

    def get_installed_pkg_keys():
        with manager_ctx.database.connect() as connection:
            return (
                connection.execute(sa.select(pkg_db.pkg.c.source, pkg_db.pkg.c.id)).tuples().all()
            )

    def make_filter_fns() -> Iterator[Callable[[cataloguer.ComputedCatalogueEntry], bool]]:
        yield lambda e: manager_ctx.config.game_flavour in e.game_flavours

        if sources:
            yield lambda e: e.source in sources

        if start_date is not None:
            start_date_ = start_date
            yield lambda e: e.last_updated >= start_date_

        if filter_installed in {'exclude', 'exclude_from_all_sources'}:
            installed_pkg_keys = set(get_installed_pkg_keys())
            if filter_installed == 'exclude_from_all_sources':
                installed_pkg_keys |= {
                    (s.source, s.id)
                    for k in installed_pkg_keys
                    for e in (catalogue.keyed_entries.get(k),)
                    if e
                    for s in e.same_as
                }

            yield lambda e: (e.source, e.id) not in installed_pkg_keys

    filter_fns = list(make_filter_fns())

    entries = catalogue.entries

    if prefer_source:
        entries = (e for e in entries if not any(s.source == prefer_source for s in e.same_as))

    if filter_installed == 'include_only':
        installed_pkg_keys = get_installed_pkg_keys()
        entries = (e for k in installed_pkg_keys for e in (catalogue.keyed_entries.get(k),) if e)

    s = _normalise_search_terms(search_terms)

    tokens_to_entries = bucketise(
        ((e.normalised_name, e) for e in entries if all(f(e) for f in filter_fns)),
        key=lambda v: v[0],
    )
    matches = rapidfuzz.process.extract(
        s,
        list(tokens_to_entries),
        scorer=rapidfuzz.fuzz.WRatio,
        limit=limit * 2,
        score_cutoff=threshold,
    )
    weighted_entries = sorted(
        (
            (-((s / 100) * ew + e.derived_download_score * dw), e)
            for m, s, _ in matches
            for _, e in tokens_to_entries[m]
        ),
        key=lambda v: v[0],
    )
    return [e for _, e in weighted_entries[:limit]]



from __future__ import annotations

import json
import typing
from collections.abc import Iterable, Iterator, Mapping, Sequence
from datetime import timedelta
from functools import reduce
from itertools import chain, product
from typing import Literal

from attrs import frozen
from cattrs import Converter
from loguru import logger
from typing_extensions import NotRequired as N
from typing_extensions import TypeAlias, TypedDict
from yarl import URL

from .http import CACHE_INDEFINITELY, make_generic_progress_ctx
from .manager_ctx import ManagerCtx
from .utils import StrEnum, bucketise, gather, read_resource_as_text, shasum, time_op
from .utils import run_in_thread as t

_LuaTable: TypeAlias = Mapping[str, '_LuaTable']
_Auras: TypeAlias = 'WeakAuras | Plateroos'
_Match: TypeAlias = tuple[Sequence['WeakAura | Plateroo'], '_WagoApiResponse', str]


class _WagoApiResponse(TypedDict):
    _id: str  # +   # Alphanumeric ID
    name: str  # +  # User-facing name
    slug: str  # +  # Slug if it has one; otherwise same as ``_id``
    url: str
    created: str  # ISO datetime
    modified: str  # ISO datetime
    game: str  # "classic" or xpac, e.g. "bfa"
    username: N[str]  # +  # Author username
    version: int  # +   # Version counter, incremented with every update
    # Semver auto-generated from ``version`` - for presentation only
    versionString: str
    changelog: _WagoApiResponse_Changelog  # +
    forkOf: N[str]  # Only present on forks
    regionType: N[str]  # Only present on WAs


class _WagoApiResponse_Changelog(TypedDict):
    format: N[Literal['bbcode', 'markdown']]
    text: N[str]


_CHECK_API_URL = URL('https://data.wago.io/api/check')
_IMPORT_STRING_API_URL = URL('https://data.wago.io/api/raw/encoded')

_aura_converter = Converter()
_aura_converter.register_structure_hook(URL, lambda v, _: URL(v))
_aura_converter.register_unstructure_hook(URL, str)


@frozen(kw_only=True)
class WeakAura:
    id: str
    uid: str
    parent: typing.Union[str, None] = None
    url: URL
    version: int

    @classmethod
    def from_lua_table(cls, lua_table: _LuaTable):
        url_string = lua_table.get('url')
        if url_string is not None:
            weakaura = _aura_converter.structure(lua_table, cls)
            if weakaura.url.host == 'wago.io':
                return weakaura


@frozen(kw_only=True)
class Plateroo(WeakAura):
    uid: str = ''

    @classmethod
    def from_lua_table(cls, lua_table: _LuaTable):
        return super().from_lua_table({**lua_table, 'id': lua_table['Name']})


@frozen
class WeakAuras:
    api_ep = 'weakauras'
    addon_name = 'WeakAuras'

    root: dict[str, list[WeakAura]]

    @classmethod
    def from_lua_table(cls, lua_table: _LuaTable):
        raw_auras = lua_table['WeakAurasSaved']['displays']
        auras = (WeakAura.from_lua_table(t) for t in raw_auras.values())
        sorted_auras = sorted(filter(None, auras), key=lambda a: a.id)
        return cls(bucketise(sorted_auras, key=lambda a: a.url.parts[1]))


@frozen
class Plateroos:
    api_ep = 'plater'
    addon_name = 'Plater'

    root: dict[str, list[Plateroo]]

    @classmethod
    def from_lua_table(cls, lua_table: _LuaTable):
        raw_auras = lua_table['PlaterDB']['profiles']
        auras = (
            Plateroo.from_lua_table(t)
            for n, p in raw_auras.items()
            for t in chain(
                ({**p, 'Name': f'__profile_{n}__'},),
                p.get('script_data') or (),
                p.get('hook_data') or (),
            )
        )
        sorted_auras = sorted(filter(None, auras), key=lambda a: a.id)
        return cls({a.url.parts[1]: [a] for a in sorted_auras})


def _merge_auras(auras: Iterable[_Auras]):
    return {
        t: t(reduce(lambda a, b: a | b, (i.root for i in a)))
        for t, a in bucketise(auras, key=type).items()
    }


class _TocNumber(StrEnum):
    Retail = '100100'
    VanillaClassic = '11403'
    Classic = '30401'


class WaCompanionBuilder:
    """A WeakAuras Companion port for shellfolk."""

    def __init__(self, manager_ctx: ManagerCtx) -> None:
        self._manager_ctx = manager_ctx

        output_folder = self._manager_ctx.config.plugins_dir / __name__
        self.addon_zip_path = output_folder / 'WeakAurasCompanion.zip'
        self.changelog_path = output_folder / 'CHANGELOG.md'
        self.version_txt_path = output_folder / 'version.txt'

    def _make_request_headers(self):
        access_token = self._manager_ctx.config.global_config.access_tokens.wago
        if access_token:
            return {'api-key': access_token}

    @staticmethod
    def extract_auras(model: type[_Auras], source: str) -> _Auras:
        from ._custom_slpp import loads

        return model.from_lua_table(loads(f'{{{source}}}'))

    def extract_installed_auras(self) -> Iterator[_Auras]:
        flavour_root = self._manager_ctx.config.addon_dir.parents[1]
        saved_vars_by_account = flavour_root.glob('WTF/Account/*/SavedVariables')
        for saved_vars, model in product(
            saved_vars_by_account,
            [WeakAuras, Plateroos],
        ):
            file = (saved_vars / model.addon_name).with_suffix('.lua')
            if not file.exists():
                logger.info(f'{file} not found')
            else:
                content = file.read_text(encoding='utf-8-sig', errors='replace')
                aura_group_cache = self._manager_ctx.config.global_config.cache_dir / shasum(
                    content
                )
                if aura_group_cache.exists():
                    logger.info(f'loading {file} from cache at {aura_group_cache}')
                    aura_group_json = json.loads(aura_group_cache.read_bytes())
                    aura_group = _aura_converter.structure({'root': aura_group_json}, model)
                else:
                    with time_op(
                        lambda t: logger.debug(
                            f'extracted {model.__name__} in {t:.3f}s'  # noqa: B023
                        )
                    ):
                        aura_group = self.extract_auras(model, content)
                    aura_group_cache.write_text(
                        json.dumps(_aura_converter.unstructure(aura_group.root)),
                        encoding='utf-8',
                    )
                yield aura_group

    async def _fetch_wago_metadata(self, api_ep: str, aura_ids: Iterable[str]):
        async with self._manager_ctx.web_client.get(
            (_CHECK_API_URL / api_ep).with_query(ids=','.join(aura_ids)),
            expire_after=timedelta(minutes=30),
            headers=self._make_request_headers(),
            trace_request_ctx=make_generic_progress_ctx('Fetching aura metadata'),
        ) as response:
            metadata: list[_WagoApiResponse]
            if response.status == 404:
                metadata = []
                return metadata

            response.raise_for_status()

            metadata = await response.json()
            return sorted(metadata, key=lambda r: r['slug'])

    async def _fetch_wago_import_string(self, remote_aura: _WagoApiResponse):
        async with self._manager_ctx.web_client.get(
            _IMPORT_STRING_API_URL.with_query(id=remote_aura['_id']).with_fragment(
                str(remote_aura['version'])
            ),
            expire_after=CACHE_INDEFINITELY,
            headers=self._make_request_headers(),
            raise_for_status=True,
            trace_request_ctx=make_generic_progress_ctx(f"Fetching aura '{remote_aura['slug']}'"),
        ) as response:
            return await response.text()

    async def get_remote_auras(self, aura_group: _Auras) -> list[_Match]:
        if not aura_group.root:
            return []

        metadata = await self._fetch_wago_metadata(aura_group.api_ep, aura_group.root)
        import_strings = await gather(self._fetch_wago_import_string(r) for r in metadata)
        return [
            (aura_group.root.get(r['slug']) or aura_group.root[r['_id']], r, i)
            for r, i in zip(metadata, import_strings)
        ]

    def _generate_addon(self, auras: Iterable[tuple[type[_Auras], list[_Match]]]):
        from zipfile import ZipFile, ZipInfo

        from mako.template import Template

        from . import _wa_templates

        def render_template(filename: str, ctx: dict[str, object]):
            return Template(read_resource_as_text(_wa_templates, filename)).render(**ctx)

        aura_dict = dict.fromkeys((WeakAuras, Plateroos), list[_Match]()) | dict(auras)

        self.addon_zip_path.parent.mkdir(exist_ok=True)
        with ZipFile(self.addon_zip_path, 'w') as file:

            def write_tpl(filename: str, ctx: dict[str, object]):
                # Not using a plain string as the first argument to ``writestr``
                # 'cause the timestamp would be set to the current time
                # which would render the build unreproducible
                zip_info = ZipInfo(filename=f'WeakAurasCompanion/{filename}')
                output = render_template(filename, ctx)
                file.writestr(zip_info, output)
                return output

            data_output = write_tpl(
                'data.lua',
                {
                    'addons': {
                        c.addon_name: [
                            (
                                metadata['slug'],
                                {
                                    'name': metadata['name'],
                                    'author': metadata.get('username', '__unknown__'),
                                    'encoded': import_string,
                                    'wagoVersion': metadata['version'],
                                    # ``wagoSemver`` is supposed to be the ``versionString``
                                    # from Wago but there is a bug where the ``version``
                                    # is sometimes not appended to the semver.
                                    # The Companion add-on's version is derived from its checksum
                                    # so if ``wagoSemver`` were to change between requests
                                    # we'd be triggering spurious updates in instawow.
                                    'wagoSemver': metadata['version'],
                                    'versionNote': metadata['changelog'].get('text', ''),
                                    'source': 'Wago',
                                },
                            )
                            for _, metadata, import_string in v
                        ]
                        for c, v in aura_dict.items()
                    },
                },
            )
            init_output = write_tpl('init.lua', {})

            interface_version = self._manager_ctx.config.game_flavour.to_flavour_keyed_enum(
                _TocNumber
            ).value
            addon_version = shasum(data_output, init_output, interface_version)[:7]

            write_tpl(
                'WeakAurasCompanion.toc',
                {'interface': interface_version, 'version': addon_version},
            )

        self.changelog_path.write_text(
            render_template(
                self.changelog_path.name,
                {
                    'changelog_entries': [
                        (
                            a.id,
                            a.url.parent,
                            metadata['version'],
                            metadata['changelog'].get('text') or 'n/a',
                        )
                        for v in aura_dict.values()
                        for existing_auras, metadata, _ in v
                        for a in (
                            next((i for i in existing_auras if not i.parent), existing_auras[0]),
                        )
                        if a.version != metadata['version']
                    ]
                },
            )
            or 'n/a',
            encoding='utf-8',
        )

        self.version_txt_path.write_text(
            addon_version,
            encoding='utf-8',
        )

    async def build(self) -> None:
        installed_auras = await t(list[_Auras])(self.extract_installed_auras())
        installed_auras_by_type = _merge_auras(installed_auras)
        aura_groups = await gather(
            self.get_remote_auras(r) for r in installed_auras_by_type.values()
        )
        await t(self._generate_addon)(zip(installed_auras_by_type, aura_groups))

    def get_version(self) -> str:
        return self.version_txt_path.read_text(encoding='utf-8')



from __future__ import annotations

import asyncio
import importlib.resources
import os
import posixpath
import string
import sys
import time
import urllib.parse
from collections import defaultdict
from collections.abc import (
    Awaitable,
    Callable,
    Hashable,
    Iterable,
    Iterator,
    Sequence,
    Set,
)
from contextlib import contextmanager
from functools import partial, wraps
from itertools import chain, groupby, islice, repeat
from pathlib import Path, PurePath
from shutil import move
from tempfile import mkdtemp
from types import ModuleType
from typing import Any, TypeVar, overload
from weakref import WeakValueDictionary

from typing_extensions import ParamSpec

_T = TypeVar('_T')
_U = TypeVar('_U')
_THashable = TypeVar('_THashable', bound=Hashable)
_P = ParamSpec('_P')


if sys.version_info >= (3, 11):
    from enum import StrEnum as StrEnum
else:
    from enum import Enum

    class StrEnum(str, Enum):
        @staticmethod
        def _generate_next_value_(name: str, start: int, count: int, last_values: list[object]):
            return name.lower()


def read_resource_as_text(package: ModuleType, resource: str, encoding: str = 'utf-8') -> str:
    return importlib.resources.files(package).joinpath(resource).read_text(encoding)


class TocReader:
    """Extracts key-value pairs from TOC files."""

    def __init__(self, contents: str) -> None:
        self.entries = {
            k: v
            for e in contents.splitlines()
            if e.startswith('##')
            for k, v in (map(str.strip, e.lstrip('#').partition(':')[::2]),)
            if k
        }

    def __getitem__(self, key: str | tuple[str, ...]) -> str | None:
        if isinstance(key, tuple):
            return next(filter(None, map(self.entries.get, key)), None)
        else:
            return self.entries.get(key)

    @classmethod
    def from_path(cls, path: Path) -> TocReader:
        return cls(path.read_text(encoding='utf-8-sig', errors='replace'))


def fill(it: Iterable[_T], fill: _T, length: int) -> Iterable[_T]:
    "Fill an iterable of specified length."
    return islice(chain(it, repeat(fill)), 0, length)


def bucketise(iterable: Iterable[_U], key: Callable[[_U], _T]) -> dict[_T, list[_U]]:
    "Place the elements of an iterable in a bucket according to ``key``."
    bucket = defaultdict[_T, list[_U]](list)
    for value in iterable:
        bucket[key(value)].append(value)
    return dict(bucket)


def chain_dict(
    keys: Iterable[_T], default: _U, *overrides: Iterable[tuple[_T, _U]]
) -> dict[_T, _U]:
    "Construct a dictionary from a series of two-tuple iterables with overlapping keys."
    return dict(chain(zip(keys, repeat(default)), *overrides))


def uniq(it: Iterable[_THashable]) -> list[_THashable]:
    "Deduplicate hashable items in an iterable maintaining insertion order."
    return list(dict.fromkeys(it))


def all_eq(it: Iterable[object]) -> bool:
    "Check that all elements of an iterable are equal."
    groups = groupby(it)
    return next(groups, True) and not next(groups, False)


def merge_intersecting_sets(it: Iterable[frozenset[_T]]) -> Iterator[frozenset[_T]]:
    "Recursively merge intersecting sets in a collection."
    many_sets = list(it)
    while many_sets:
        this_set = many_sets.pop(0)
        while True:
            for idx, other_set in enumerate(many_sets):
                if not this_set.isdisjoint(other_set):
                    this_set |= many_sets.pop(idx)
                    break
            else:
                break
        yield this_set


@overload
async def gather(it: Iterable[Awaitable[_U]], wrapper: None = None) -> list[_U]:
    ...


@overload
async def gather(
    it: Iterable[Awaitable[_U]],
    wrapper: Callable[[Awaitable[_U]], Awaitable[_T]],
) -> list[_T]:
    ...


async def gather(
    it: Iterable[Awaitable[object]], wrapper: Callable[..., Awaitable[object]] | None = None
) -> Sequence[object]:
    if wrapper is not None:
        it = map(wrapper, it)
    return await asyncio.gather(*it)


def run_in_thread(fn: Callable[_P, _U]) -> Callable[_P, Awaitable[_U]]:
    @wraps(fn)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs):
        loop = asyncio.get_running_loop()
        return loop.run_in_executor(None, partial(fn, *args, **kwargs))

    return wrapper


def tabulate(rows: Sequence[tuple[object, ...]], *, max_col_width: int = 60) -> str:
    "Produce an ASCII table from equal-length elements in a sequence."
    from textwrap import fill

    def apply_max_col_width(value: object):
        return fill(str(value), width=max_col_width, max_lines=1)

    def calc_resultant_col_widths(rows: Sequence[tuple[str, ...]]):
        cols = zip(*rows)
        return [max(map(len, c)) for c in cols]

    norm_rows = [tuple(apply_max_col_width(i) for i in r) for r in rows]
    head, *tail = norm_rows

    base_template = '  '.join(f'{{{{{{0}}{w}}}}}' for w in calc_resultant_col_widths(norm_rows))
    row_template = base_template.format(':<')
    table = '\n'.join(
        (
            base_template.format(':^').format(*head),
            base_template.format('0:-<').format(''),
            *(row_template.format(*r) for r in tail),
        )
    )
    return table


def trash(paths: Iterable[PurePath], *, dest: PurePath, missing_ok: bool = False) -> None:
    paths_iter = iter(paths)
    first_path = next(paths_iter, None)

    if first_path is None:
        return

    exc_classes = FileNotFoundError if missing_ok else ()

    parent_folder = mkdtemp(dir=dest, prefix=f'deleted-{first_path.name}-')

    for path in chain((first_path,), paths_iter):
        try:
            move(path, parent_folder)
        except exc_classes:
            pass


def shasum(*values: object) -> str:
    "Base-16-encode a string using SHA-256 truncated to 32 characters."
    from hashlib import sha256

    return sha256(''.join(map(str, filter(None, values))).encode()).hexdigest()[:32]


def find_addon_zip_tocs(names: Iterable[str]) -> Iterator[tuple[str, str]]:
    "Find top-level folders in a list of ZIP member paths."
    for name in names:
        if name.count(posixpath.sep) == 1:
            head, tail = posixpath.split(name)
            if tail.startswith(head) and tail[-4:].lower() == '.toc':
                yield (name, head)


def make_zip_member_filter_fn(base_dirs: Set[str]) -> Callable[[str], bool]:
    "Filter out items which are not sub-paths of top-level folders in a ZIP."

    def is_subpath(name: str):
        head, sep, _ = name.partition(posixpath.sep)
        return head in base_dirs if sep else False

    return is_subpath


def is_file_uri(uri: str) -> bool:
    return uri.startswith('file://')


def file_uri_to_path(file_uri: str) -> str:
    "Convert a file URI to a path that works both on Windows and *nix."
    from urllib.parse import unquote

    unprefixed_path = unquote(file_uri.removeprefix('file://'))
    # A slash is prepended to the path even when there isn't one there
    # on Windows.  The ``PurePath`` instance will inherit from either
    # ``PurePosixPath`` or ``PureWindowsPath``; this will be a no-op on POSIX.
    if PurePath(unprefixed_path[1:]).drive:
        unprefixed_path = unprefixed_path[1:]
    return unprefixed_path


def as_plain_text_data_url(body: str = '') -> str:
    return f'data:,{urllib.parse.quote(body)}'


def extract_byte_range_offset(content_range: str):
    return int(content_range.replace('bytes ', '').partition('-')[0])


def normalise_names(replace_delim: str) -> Callable[[str], str]:
    trans_table = str.maketrans(dict.fromkeys(string.punctuation, ' '))

    def normalise(value: str):
        return replace_delim.join(value.casefold().translate(trans_table).split())

    return normalise


slugify = normalise_names('-')


def reveal_folder(path: str | os.PathLike[str]) -> None:
    if sys.platform == 'win32':
        os.startfile(path, 'explore')
    else:
        import click

        click.launch(os.fspath(path), locate=True)


@contextmanager
def time_op(on_complete: Callable[[float], None]) -> Iterator[None]:
    start = time.perf_counter()
    yield
    on_complete(time.perf_counter() - start)


class WeakValueDefaultDictionary(WeakValueDictionary[_T, _U]):
    def __init__(self, default_factory: Callable[[], _U]) -> None:
        super().__init__()
        self.__default_factory = default_factory

    def __getitem__(self, key: _T) -> _U:
        try:
            return super().__getitem__(key)
        except KeyError:
            default = self[key] = self.__default_factory()
            return default


def add_exc_note(exc: BaseException, note: str) -> None:
    if sys.version_info >= (3, 11):
        exc.add_note(note)

    else:
        exc_without_notes: Any = exc

        if not hasattr(exc, '__notes__'):
            exc_without_notes.__notes__ = list[str]()
        exc_without_notes.__notes__.append(note)



from __future__ import annotations

import enum
import urllib.parse
from collections.abc import AsyncIterator, Collection, Iterable, Sequence
from functools import update_wrapper
from pathlib import Path
from typing import Any, ClassVar, Protocol, TypeVar

from typing_extensions import Self
from yarl import URL

from . import http, manager_ctx, pkg_models
from . import results as R
from .catalogue.cataloguer import CatalogueEntry
from .common import AddonHashMethod, Defn, SourceMetadata
from .config import GlobalConfig
from .utils import file_uri_to_path, gather, run_in_thread


class FolderHashCandidate(Protocol):  # pragma: no cover
    name: str

    def hash_contents(self, __method: AddonHashMethod) -> str:
        ...


TFolderHashCandidate = TypeVar('TFolderHashCandidate', bound=FolderHashCandidate)


class HeadersIntent(enum.IntEnum):
    Download = enum.auto()


class Resolver(Protocol):  # pragma: no cover
    metadata: ClassVar[SourceMetadata]
    requires_access_token: ClassVar[str | None]

    def __init__(self, manager_ctx: manager_ctx.ManagerCtx) -> None:
        ...

    @classmethod
    def get_alias_from_url(cls, url: URL) -> str | None:
        "Attempt to extract a ``Defn`` alias from a given URL."
        ...

    async def make_request_headers(
        self, intent: HeadersIntent | None = None
    ) -> dict[str, str] | None:
        "Create headers for resolver HTTP requests."
        ...

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, pkg_models.Pkg | R.ManagerError | R.InternalError]:
        "Resolve multiple ``Defn``s into packages."
        ...

    async def resolve_one(self, defn: Defn, metadata: Any) -> pkg_models.Pkg:
        "Resolve a ``Defn`` into a package."
        ...

    async def get_changelog(self, uri: URL) -> str:
        "Retrieve a changelog from a URI."
        ...

    async def get_folder_hash_matches(
        self, candidates: Collection[TFolderHashCandidate]
    ) -> Iterable[tuple[Defn, frozenset[TFolderHashCandidate]]]:
        "Find ``Defn``s from folder fingerprint."
        ...

    @classmethod
    def catalogue(cls, web_client: http.ClientSessionType) -> AsyncIterator[CatalogueEntry]:
        "Enumerate add-ons from the source."
        ...


class BaseResolver(Resolver, Protocol):
    _manager_ctx: manager_ctx.ManagerCtx

    def __init__(self, manager_ctx: manager_ctx.ManagerCtx) -> None:
        self._manager_ctx = manager_ctx

    __orig_init = __init__

    def __init_subclass__(cls) -> None:
        # ``Protocol`` clobbers ``__init__`` in Python < 3.11.  The fix was
        # also backported to 3.9 and 3.10 at some point.
        if cls.__init__ is _DummyResolver.__init__:
            cls.__init__ = cls.__orig_init

        if cls.resolve_one is not super().resolve_one:

            async def resolve_one(self: Self, defn: Defn, metadata: Any) -> pkg_models.Pkg:
                extraneous_strategies = (
                    defn.strategies.filled_strategies.keys() - self.metadata.strategies
                )
                if extraneous_strategies:
                    raise R.PkgStrategiesUnsupported(extraneous_strategies)
                return await cls_resolve_one(self, defn, metadata)

            cls_resolve_one = cls.resolve_one
            setattr(cls, cls.resolve_one.__name__, update_wrapper(resolve_one, cls.resolve_one))

    @classmethod
    def _get_access_token(cls, global_config: GlobalConfig, name: str | None = None) -> str | None:
        name = name or cls.requires_access_token
        if name is not None:
            return getattr(global_config.access_tokens, name, None)

    @classmethod
    def get_alias_from_url(cls, url: URL) -> str | None:
        return None

    async def make_request_headers(
        self, intent: HeadersIntent | None = None
    ) -> dict[str, str] | None:
        return None

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, pkg_models.Pkg | R.ManagerError | R.InternalError]:
        results = await gather((self.resolve_one(d, None) for d in defns), R.resultify_async_exc)
        return dict(zip(defns, results))

    async def get_changelog(self, uri: URL) -> str:
        if uri.scheme == 'data' and uri.raw_path.startswith(','):
            return urllib.parse.unquote(uri.raw_path[1:])
        elif uri.scheme in {'http', 'https'}:
            async with self._manager_ctx.web_client.get(
                uri,
                expire_after=http.CACHE_INDEFINITELY,
                headers=await self.make_request_headers(),
                raise_for_status=True,
            ) as response:
                return await response.text()
        elif uri.scheme == 'file':
            return await run_in_thread(Path(file_uri_to_path(str(uri))).read_text)(
                encoding='utf-8'
            )
        else:
            raise ValueError('Unsupported URI with scheme', uri.scheme)

    async def get_folder_hash_matches(
        self, candidates: Collection[TFolderHashCandidate]
    ) -> Iterable[tuple[Defn, frozenset[TFolderHashCandidate]]]:
        return []

    @classmethod
    async def catalogue(cls, web_client: http.ClientSessionType) -> AsyncIterator[CatalogueEntry]:
        return
        yield


class _DummyResolver(Resolver):  # pyright: ignore  # noqa: PGH003
    pass



"""
Add cascade deletes to the pkg table's dependents.

Revision ID: e13430219249
Revises: 75f69831f74f
Create Date: 2022-04-30 22:29:49.359123

"""
from __future__ import annotations

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'e13430219249'
down_revision = '75f69831f74f'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table(
        'pkg_dep',
        schema=None,
        reflect_args=[
            sa.ForeignKeyConstraint(
                ['pkg_source', 'pkg_id'],
                ['pkg.source', 'pkg.id'],
                name='fk_pkg_dep_pkg_source_and_id',
            ),
        ],
    ) as batch_op:
        batch_op.drop_constraint('fk_pkg_dep_pkg_source_and_id', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_dep_pkg_source_and_id',
            'pkg',
            ['pkg_source', 'pkg_id'],
            ['source', 'id'],
            ondelete='CASCADE',
        )

    with op.batch_alter_table(
        'pkg_folder',
        schema=None,
        reflect_args=[
            sa.ForeignKeyConstraint(
                ['pkg_source', 'pkg_id'],
                ['pkg.source', 'pkg.id'],
                name='fk_pkg_folder_pkg_source_and_id',
            ),
        ],
    ) as batch_op:
        batch_op.drop_constraint('fk_pkg_folder_pkg_source_and_id', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_folder_pkg_source_and_id',
            'pkg',
            ['pkg_source', 'pkg_id'],
            ['source', 'id'],
            ondelete='CASCADE',
        )

    with op.batch_alter_table(
        'pkg_options',
        schema=None,
        reflect_args=[
            sa.ForeignKeyConstraint(
                ['pkg_source', 'pkg_id'],
                ['pkg.source', 'pkg.id'],
                name='fk_pkg_options_pkg_source_and_id',
            ),
        ],
    ) as batch_op:
        batch_op.drop_constraint('fk_pkg_options_pkg_source_and_id', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_options_pkg_source_and_id',
            'pkg',
            ['pkg_source', 'pkg_id'],
            ['source', 'id'],
            ondelete='CASCADE',
        )

    with op.batch_alter_table(
        'pkg_version_log',
        schema=None,
        reflect_args=[
            sa.ForeignKeyConstraint(
                ['pkg_source', 'pkg_id'],
                ['pkg.source', 'pkg.id'],
                name='fk_pkg_version_log_pkg_source_and_id',
            ),
        ],
    ) as batch_op:
        batch_op.drop_constraint('fk_pkg_version_log_pkg_source_and_id', type_='foreignkey')


def downgrade():
    with op.batch_alter_table('pkg_options', schema=None) as batch_op:
        batch_op.drop_constraint('fk_pkg_options_pkg_source_and_id', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_options_pkg_source_and_id', 'pkg', ['pkg_source', 'pkg_id'], ['source', 'id']
        )

    with op.batch_alter_table('pkg_folder', schema=None) as batch_op:
        batch_op.drop_constraint('fk_pkg_folder_pkg_source_and_id', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_folder_pkg_source_and_id', 'pkg', ['pkg_source', 'pkg_id'], ['source', 'id']
        )

    with op.batch_alter_table('pkg_dep', schema=None) as batch_op:
        batch_op.drop_constraint('fk_pkg_dep_pkg_source_and_id', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_dep_pkg_source_and_id', 'pkg', ['pkg_source', 'pkg_id'], ['source', 'id']
        )

    with op.batch_alter_table('pkg_version_log', schema=None) as batch_op:
        batch_op.create_foreign_key(
            'fk_pkg_version_log_pkg_source_and_id',
            'pkg',
            ['pkg_source', 'pkg_id'],
            ['source', 'id'],
        )



from __future__ import annotations

import sys
from pathlib import Path

import pytest

from instawow.common import Flavour
from instawow.wow_installations import find_installations, infer_flavour_from_addon_dir


@pytest.mark.parametrize(
    ('path', 'flavour'),
    [
        (
            'wowzerz/_classic_/Interface/AddOns',
            Flavour.Classic,
        ),
        (
            '/foo/bar/_classic_ptr_/Interface/AddOns',
            Flavour.Classic,
        ),
        (
            '_classic_era_/Interface/AddOns',
            Flavour.VanillaClassic,
        ),
        (
            '_classic_era_ptr_/Interface/AddOns',
            Flavour.VanillaClassic,
        ),
        (
            'wowzerz/_retail_/Interface/AddOns',
            Flavour.Retail,
        ),
        (
            'anything goes',
            None,
        ),
    ],
)
def test_can_infer_flavour_from_addon_dir(
    path: str,
    flavour: Flavour | None,
):
    assert infer_flavour_from_addon_dir(path) is flavour


@pytest.mark.skipif(sys.platform != 'darwin', reason='Only supported on Mac')
def test_can_find_mac_installations(
    monkeypatch: pytest.MonkeyPatch,
):
    with monkeypatch.context() as patcher:

        def check_output_no_installation(*args, **kwargs):
            return ''

        patcher.setattr('subprocess.check_output', check_output_no_installation)

        assert set(find_installations()) == set()

    with monkeypatch.context() as patcher:
        app_bundle_paths = {
            Path('/Applications/World of Warcraft/_retail_/World of Warcraft.app'): Flavour.Retail,
            Path(
                '/Applications/World of Warcraft/_classic_/World of Warcraft Classic.app'
            ): Flavour.Classic,
        }

        def check_output_has_installations(*args, **kwargs):
            return '\n'.join(map(str, app_bundle_paths))

        patcher.setattr('subprocess.check_output', check_output_has_installations)

        assert set(find_installations()) == {(p.parent, f) for p, f in app_bundle_paths.items()}



from __future__ import annotations

import datetime as dt

import pytest

from instawow.catalogue.search import search
from instawow.common import Defn, Flavour
from instawow.manager_ctx import ManagerCtx
from instawow.pkg_management import PkgManager
from instawow.results import PkgInstalled


async def test_basic_search(
    iw_manager_ctx: ManagerCtx,
):
    limit = 5
    results = await search(iw_manager_ctx, 'molinari', limit=limit)
    assert len(results) <= 5
    assert {('curse', 'molinari'), ('wowi', '13188')} <= {
        (e.source, e.slug or e.id) for e in results
    }


@pytest.mark.parametrize(
    'iw_config_values',
    Flavour,
    indirect=True,
)
async def test_search_flavour_filtering(
    iw_manager_ctx: ManagerCtx,
):
    results = await search(iw_manager_ctx, 'atlas loot classic', limit=10)
    has_atlas = ('curse', 'atlaslootclassic') in {(e.source, e.slug or e.id) for e in results}
    if iw_manager_ctx.config.game_flavour in {
        Flavour.VanillaClassic,
        Flavour.Classic,
    }:
        assert has_atlas
    else:
        assert not has_atlas


async def test_search_source_filtering(
    iw_manager_ctx: ManagerCtx,
):
    results = await search(iw_manager_ctx, 'molinari', limit=5, sources={'curse'})
    assert all(e.source == 'curse' for e in results)
    assert ('curse', 'molinari') in {(e.source, e.slug) for e in results}


async def test_search_date_filtering(
    iw_manager_ctx: ManagerCtx,
):
    start_date = dt.datetime.now(tz=dt.timezone.utc) - dt.timedelta(days=365)
    results = await search(iw_manager_ctx, 'molinari', limit=5, start_date=start_date)
    assert all(e.last_updated > start_date for e in results)


async def test_search_unknown_source(
    iw_manager_ctx: ManagerCtx,
):
    with pytest.raises(ValueError, match='Unknown source'):
        await search(iw_manager_ctx, 'molinari', limit=5, sources={'foo'})


async def test_search_filter_installed(
    iw_manager_ctx: ManagerCtx,
    iw_manager: PkgManager,
):
    results = await search(iw_manager_ctx, 'molinari', limit=5, filter_installed='include_only')
    assert not results

    defn = Defn('curse', 'molinari')

    install_result = (await iw_manager.install([defn], False))[defn]
    assert type(install_result) is PkgInstalled

    results = await search(iw_manager_ctx, 'molinari', limit=5)
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} <= {
        (e.source, e.slug) for e in results
    }

    results = await search(iw_manager_ctx, 'molinari', limit=5, filter_installed='include_only')
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} & {
        (e.source, e.slug) for e in results
    } == {('curse', 'molinari')}

    results = await search(iw_manager_ctx, 'molinari', limit=5, filter_installed='exclude')
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} & {
        (e.source, e.slug) for e in results
    } == {('github', 'p3lim-wow/molinari')}

    results = await search(
        iw_manager_ctx, 'molinari', limit=5, filter_installed='exclude_from_all_sources'
    )
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} & {
        (e.source, e.slug) for e in results
    } == set()


async def test_search_prefer_known_source(
    iw_manager_ctx: ManagerCtx,
):
    results = await search(iw_manager_ctx, 'molinari', limit=5, prefer_source=None)
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} <= {
        (e.source, e.slug) for e in results
    }

    results = await search(iw_manager_ctx, 'molinari', limit=5, prefer_source='github')
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} & {
        (e.source, e.slug) for e in results
    } == {('github', 'p3lim-wow/molinari')}


async def test_search_prefer_unknown_source(
    iw_manager_ctx: ManagerCtx,
):
    with pytest.raises(ValueError, match='Unknown preferred source: foo'):
        await search(iw_manager_ctx, 'molinari', limit=5, prefer_source='foo')



from __future__ import annotations

import sys
from pathlib import Path
from typing import Any

import pytest
from cattrs import AttributeValidationNote, ClassValidationError

from instawow.common import Flavour
from instawow.config import Config, GlobalConfig


def test_env_vars_always_take_precedence(
    monkeypatch: pytest.MonkeyPatch,
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    monkeypatch.setenv('INSTAWOW_AUTO_UPDATE_CHECK', '1')
    monkeypatch.setenv('INSTAWOW_GAME_FLAVOUR', 'classic')

    global_config = GlobalConfig.from_values(
        {**iw_global_config_values, 'auto_update_check': False}, env=True
    )
    config = Config.from_values(
        {'global_config': global_config, **iw_config_values, 'game_flavour': 'retail'}, env=True
    )
    assert global_config.auto_update_check is True
    assert config.game_flavour is Flavour.Classic


def test_read_profile_from_nonexistent_config_dir_raises(
    iw_global_config_values: dict[str, Any],
):
    global_config = GlobalConfig(config_dir=iw_global_config_values['config_dir'])
    with pytest.raises(FileNotFoundError):
        Config.read(global_config, '__default__')


def test_init_with_nonexistent_addon_dir_raises(
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    global_config = GlobalConfig.from_values(iw_global_config_values).write()
    with pytest.raises(ValueError, match='not a writable directory'):
        Config(global_config=global_config, **{**iw_config_values, 'addon_dir': '#@$foo'})


def test_default_config_dir(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.delenv('XDG_CONFIG_HOME', False)

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'linux')

        config_dir = GlobalConfig().config_dir
        assert config_dir == Path.home() / '.config/instawow'

        patcher.setenv('XDG_CONFIG_HOME', '/foo')
        config_dir = GlobalConfig().config_dir
        assert config_dir == Path('/foo/instawow').resolve()

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'darwin')

        config_dir = GlobalConfig().config_dir
        assert config_dir == Path.home() / 'Library/Application Support/instawow'

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'win32')

        patcher.delenv('APPDATA', False)
        assert GlobalConfig().config_dir == Path.home() / '.config' / 'instawow'

        patcher.setenv('APPDATA', '/foo')
        assert GlobalConfig().config_dir == Path('/foo/instawow').resolve()


def test_config_dir_xdg_env_var_is_respected_on_all_plats(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
):
    config_parent_dir = tmp_path / 'instawow_config_parent_dir'
    monkeypatch.setenv('XDG_CONFIG_HOME', str(config_parent_dir))
    config = GlobalConfig()
    assert config.config_dir == config_parent_dir / 'instawow'


def test_config_dir_instawow_specific_env_var_takes_precedence(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
):
    test_config_dir_xdg_env_var_is_respected_on_all_plats(monkeypatch, tmp_path)

    config_dir = tmp_path / 'instawow_config_dir'
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(config_dir))
    config = GlobalConfig.from_values(env=True)
    assert config.config_dir == config_dir


def test_default_state_dir(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.delenv('XDG_STATE_HOME', False)

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'linux')

        assert GlobalConfig().state_dir == Path.home() / '.local' / 'state' / 'instawow'

    for platform in {'darwin', 'win32'}:
        with monkeypatch.context() as patcher:
            patcher.setattr(sys, 'platform', platform)

            config = GlobalConfig()
            assert config.config_dir == config.state_dir


def test_state_dir_xdg_env_var_is_respected_on_all_plats(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
):
    state_parent_dir = tmp_path / 'instawow_state_parent_dir'
    monkeypatch.setenv('XDG_STATE_HOME', str(state_parent_dir))
    config = GlobalConfig()
    assert config.state_dir == state_parent_dir / 'instawow'


def test_state_dir_instawow_specific_env_var_takes_precedence(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
):
    test_state_dir_xdg_env_var_is_respected_on_all_plats(monkeypatch, tmp_path)

    state_dir = tmp_path / 'instawow_state_dir'
    monkeypatch.setenv('INSTAWOW_STATE_DIR', str(state_dir))
    config = GlobalConfig.from_values(env=True)
    assert config.state_dir == state_dir


def test_can_list_profiles(
    iw_config_values: dict[str, Any],
):
    global_config = GlobalConfig.read()
    assert global_config.list_profiles() == []

    Config.from_values({'global_config': global_config, **iw_config_values}).write()
    Config(global_config=global_config, **{**iw_config_values, 'profile': 'foo'}).write()
    assert set(global_config.list_profiles()) == {'__default__', 'foo'}


def test_profile_dirs_are_populated(
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    global_config = GlobalConfig.from_values(iw_global_config_values)
    config = Config.from_values({'global_config': global_config, **iw_config_values}).write()
    assert {i.name for i in config.config_dir.iterdir()} <= {'config.json'}
    assert {i.name for i in config.state_dir.iterdir()} <= {'logs', 'plugins'}


def test_can_delete_profile(
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    global_config = GlobalConfig.from_values(iw_global_config_values).write()
    config = Config.from_values({'global_config': global_config, **iw_config_values}).write()
    assert config.config_dir.exists()
    config.delete()
    assert not config.config_dir.exists()


def test_validate_profile_name(
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    global_config = GlobalConfig.from_values(iw_global_config_values)

    with pytest.raises(ClassValidationError) as exc_info:
        Config.from_values({'global_config': global_config, **iw_config_values, 'profile': ''})

    (value_error,) = exc_info.value.exceptions
    assert value_error.args == ('Value must have a minimum length of 1',)

    (note,) = value_error.__notes__
    assert note == 'Structuring class Config @ attribute profile'
    assert type(note) is AttributeValidationNote
    assert note.name == 'profile'


@pytest.mark.skipif(
    sys.platform == 'win32',
    reason='chmod has no effect on Windows',
)
def test_validate_addon_dir(
    tmp_path: Path,
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    global_config = GlobalConfig.from_values(iw_global_config_values)

    non_writeable_dir = tmp_path / 'non-writeable-dir'
    non_writeable_dir.mkdir(0o400)

    with pytest.raises(ClassValidationError) as exc_info:
        Config.from_values(
            {'global_config': global_config, **iw_config_values, 'addon_dir': non_writeable_dir}
        )

    (value_error,) = exc_info.value.exceptions
    assert value_error.args == (f'"{non_writeable_dir}" is not a writable directory',)

    (note,) = value_error.__notes__
    assert note == 'Structuring class Config @ attribute addon_dir'
    assert type(note) is AttributeValidationNote
    assert note.name == 'addon_dir'



from __future__ import annotations

import os
import subprocess
import sys
from collections.abc import Iterator
from pathlib import Path, PurePath

from typing_extensions import TypeAlias

from .common import Flavour

_InstallationPath: TypeAlias = Path

_DELECTABLE_DIR_NAMES = {
    '_retail_': Flavour.Retail,
    '_ptr_': Flavour.Retail,
    '_xptr_': Flavour.Retail,
    '_classic_era_': Flavour.VanillaClassic,
    '_classic_era_ptr_': Flavour.VanillaClassic,
    '_classic_': Flavour.Classic,
    '_classic_ptr_': Flavour.Classic,
}

ADDON_DIR_PARTS = ('Interface', 'AddOns')
_NORMALISED_ADDON_DIR_PARTS = tuple(map(str.casefold, ADDON_DIR_PARTS))


def infer_flavour_from_addon_dir(path: os.PathLike[str] | str) -> Flavour | None:
    tail = tuple(map(str.casefold, PurePath(path).parts[-3:]))
    if len(tail) == 3 and tail[1:] == _NORMALISED_ADDON_DIR_PARTS:
        return _DELECTABLE_DIR_NAMES.get(tail[0])


def _find_mac_installations():
    try:
        possible_installations = subprocess.check_output(
            [
                'mdfind',
                "kMDItemCFBundleIdentifier == 'com.blizzard.worldofwarcraft'",
            ],
            text=True,
            timeout=3,
        ).splitlines()
    except subprocess.TimeoutExpired:
        pass
    else:
        for match in possible_installations:
            installation_path = Path(match).parent
            yield (installation_path, _DELECTABLE_DIR_NAMES.get(installation_path.name))


def find_installations() -> Iterator[tuple[_InstallationPath, Flavour | None]]:
    if sys.platform == 'darwin':
        yield from _find_mac_installations()



from __future__ import annotations

from typing import Any

from instawow.common import ChangelogFormat, Defn, SourceMetadata, Strategy, StrategyValues
from instawow.manager_ctx import ManagerCtx
from instawow.resolvers import BaseResolver
from instawow.results import PkgStrategiesUnsupported


async def test_unsupported_strategies_raise(
    iw_manager_ctx: ManagerCtx,
):
    class Resolver(BaseResolver):
        metadata = SourceMetadata(
            id='foo',
            name='Foo',
            strategies=frozenset(),
            changelog_format=ChangelogFormat.Raw,
            addon_toc_key=None,
        )

        def resolve_one(self, defn: Defn, metadata: Any):
            raise NotImplementedError

    defn = Defn(
        Resolver.metadata.id,
        'foo',
        strategies=StrategyValues(
            any_flavour=True,
            any_release_type=True,
            version_eq='0',
        ),
    )

    result = (await Resolver(iw_manager_ctx).resolve([defn]))[defn]

    assert type(result) is PkgStrategiesUnsupported
    assert (
        result.message
        == f'strategies are not valid for source: {Strategy.AnyFlavour}, {Strategy.AnyReleaseType}, {Strategy.VersionEq}'
    )



from __future__ import annotations

from datetime import datetime, timezone

import click

import instawow.plugins
from instawow.common import ChangelogFormat, Defn, SourceMetadata
from instawow.pkg_models import Pkg, PkgOptions
from instawow.resolvers import BaseResolver


@click.command
def foo():
    "don't foo where you bar"
    print('success!')


class MyResolver(BaseResolver):
    metadata = SourceMetadata(
        id='me',
        name="It's me",
        strategies=frozenset(),
        changelog_format=ChangelogFormat.Markdown,
        addon_toc_key=None,
    )
    requires_access_token = None

    async def resolve_one(self, defn: Defn, metadata: None) -> Pkg:
        return Pkg(
            source=self.metadata.id,
            id='bar',
            slug='bar',
            name='Bar',
            description='The quintessential bar add-on, brought to you by yours truly',
            url='http://example.com/',
            download_url='file:///...',
            date_published=datetime.now(timezone.utc),
            version='0',
            changelog_url='data:,',
            options=PkgOptions.from_strategy_values(defn.strategies),
        )


@instawow.plugins.hookimpl
def instawow_add_commands():
    return (foo,)


@instawow.plugins.hookimpl
def instawow_add_resolvers():
    return (MyResolver,)



from __future__ import annotations

import re

import pytest
from typing_extensions import assert_never
from yarl import URL

from instawow._sources.cfcore import CfCoreResolver
from instawow.common import Defn, Flavour, Strategy, StrategyValues
from instawow.manager_ctx import ManagerCtx
from instawow.pkg_management import PkgManager
from instawow.pkg_models import Pkg
from instawow.results import PkgFilesNotMatching

CURSE_IDS = {
    'big-wigs': '2382',
    'molinari': '20338',
    'adibags': '23350',
    'mythic-dungeon-tools': '288981',
    'classiccastbars': '322865',
    'elkbuffbars': '2398',
    'atlaslootclassic': '326516',
    'elvui-adibags': '333072',
    'bigwigs-voice-korean': '402180',
}


@pytest.fixture
def curse_resolver(
    iw_manager_ctx: ManagerCtx,
):
    return CfCoreResolver(iw_manager_ctx)


@pytest.mark.parametrize(
    'iw_config_values',
    Flavour,
    indirect=True,
)
async def test_resolve_flavourful_addon(
    curse_resolver: CfCoreResolver,
):
    defn = Defn('curse', CURSE_IDS['classiccastbars'])

    result = (await curse_resolver.resolve([defn]))[defn]
    assert type(result) is Pkg


@pytest.mark.parametrize(
    'iw_config_values',
    Flavour,
    indirect=True,
)
async def test_resolve_classic_onlu_addon(
    iw_manager_ctx: ManagerCtx,
    curse_resolver: CfCoreResolver,
):
    defn = Defn('curse', CURSE_IDS['atlaslootclassic'])

    result = (await curse_resolver.resolve([defn]))[defn]

    if (
        iw_manager_ctx.config.game_flavour is Flavour.VanillaClassic
        or iw_manager_ctx.config.game_flavour is Flavour.Classic
    ):
        assert type(result) is Pkg
    elif iw_manager_ctx.config.game_flavour is Flavour.Retail:
        assert type(result) is PkgFilesNotMatching
        assert (
            result.message
            == f'no files found for: {Strategy.AnyFlavour}=None; {Strategy.AnyReleaseType}=None; {Strategy.VersionEq}=None'
        )
    else:
        assert_never(iw_manager_ctx.config.game_flavour)


async def test_curse_any_flavour_strategy(
    curse_resolver: CfCoreResolver,
):
    flavourful = Defn(
        'curse', CURSE_IDS['classiccastbars'], strategies=StrategyValues(any_flavour=True)
    )
    classics_only = Defn(
        'curse', CURSE_IDS['atlaslootclassic'], strategies=StrategyValues(any_flavour=True)
    )

    results = await curse_resolver.resolve([flavourful, classics_only])
    assert all(type(r) is Pkg for r in results.values())


async def test_curse_slug_match(
    curse_resolver: CfCoreResolver,
):
    defn = Defn('curse', 'molinari')

    result = (await curse_resolver.resolve([defn]))[defn]
    assert type(result) is Pkg
    assert result.id == CURSE_IDS['molinari']


async def test_curse_version_pinning(
    curse_resolver: CfCoreResolver,
):
    defn = Defn('curse', 'molinari', strategies=StrategyValues(version_eq='100005.97-Release'))

    result = (await curse_resolver.resolve([defn]))[defn]
    assert type(result) is Pkg
    assert result.options.version_eq is True
    assert result.version == '100005.97-Release'


async def test_curse_deps_retrieved(
    iw_manager: PkgManager,
):
    defn = Defn('curse', CURSE_IDS['bigwigs-voice-korean'])

    results = await iw_manager.resolve([defn], with_deps=True)
    assert all(type(r) is Pkg for r in results.values())
    assert {'bigwigs-voice-korean', 'big-wigs'} == {p.slug for p in results.values()}


async def test_changelog_url_format(
    curse_resolver: CfCoreResolver,
):
    defn = Defn('curse', CURSE_IDS['classiccastbars'])

    result = (await curse_resolver.resolve([defn]))[defn]
    assert type(result) is Pkg
    assert re.match(
        r'https://api\.curseforge\.com/v1/mods/\d+/files/\d+/changelog',
        result.changelog_url,
    )


@pytest.mark.parametrize(
    ('url', 'extracted_alias'),
    [
        (
            'https://www.curseforge.com/wow/addons/molinari',
            'molinari',
        ),
        (
            'https://www.curseforge.com/wow/addons/molinari/download',
            'molinari',
        ),
    ],
)
def test_can_extract_alias_from_url(
    curse_resolver: CfCoreResolver,
    url: str,
    extracted_alias: str,
):
    assert curse_resolver.get_alias_from_url(URL(url)) == extracted_alias



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from functools import lru_cache, partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

from loguru import logger
from typing_extensions import TypeAlias, TypedDict

from . import common
from .utils import read_resource_as_text

if TYPE_CHECKING:
    import aiohttp

    ClientSessionType: TypeAlias = aiohttp.ClientSession
else:
    ClientSessionType = type

_USER_AGENT = 'instawow (+https://github.com/layday/instawow)'

_DEFAULT_EXPIRE = 0  # Do not cache by default.

CACHE_INDEFINITELY = -1


class _GenericDownloadTraceRequestCtx(TypedDict):
    report_progress: Literal['generic']
    label: str


class _DefnDownloadTraceRequestCtx(TypedDict):
    report_progress: Literal['pkg_download']
    profile: str
    defn: common.Defn


TraceRequestCtx: TypeAlias = (
    '_GenericDownloadTraceRequestCtx | _DefnDownloadTraceRequestCtx | None'
)


def make_generic_progress_ctx(label: str) -> _GenericDownloadTraceRequestCtx:
    return {'report_progress': 'generic', 'label': label}


def make_defn_progress_ctx(profile: str, defn: common.Defn) -> _DefnDownloadTraceRequestCtx:
    return {'report_progress': 'pkg_download', 'profile': profile, 'defn': defn}


@lru_cache(1)
def _load_certifi_certs():
    try:
        import certifi
    except ModuleNotFoundError:
        pass
    else:
        logger.info('loading certifi certs')
        return read_resource_as_text(certifi, 'cacert.pem', encoding='ascii')


@asynccontextmanager
async def init_web_client(
    cache_dir: Path | None, *, no_cache: bool = False, **kwargs: Any
) -> AsyncIterator[ClientSessionType]:
    from aiohttp import ClientSession, ClientTimeout, TCPConnector

    make_connector = partial(TCPConnector, limit_per_host=10)
    certifi_certs = _load_certifi_certs()
    if certifi_certs:
        import ssl

        make_connector = partial(
            make_connector, ssl=ssl.create_default_context(cadata=certifi_certs)
        )

    kwargs = {
        'connector': make_connector(),
        'headers': {'User-Agent': _USER_AGENT},
        'trust_env': True,  # Respect the ``http(s)_proxy`` env var
        'timeout': ClientTimeout(connect=60, sock_connect=10, sock_read=20),
        **kwargs,
    }

    if cache_dir is not None:
        from aiohttp_client_cache.session import CachedSession

        from ._http_cache_db import SQLiteBackend, acquire_cache_connection

        with acquire_cache_connection(cache_dir) as connection:
            cache_backend = SQLiteBackend(
                allowed_codes=[200, 206],
                allowed_methods=['GET', 'POST'],
                connection=connection,
                expire_after=_DEFAULT_EXPIRE,
                include_headers=True,
            )
            if no_cache:
                cache_backend.disabled = True

            async with CachedSession(
                cache=cache_backend,
                **kwargs,
            ) as client_session:
                yield client_session
    else:
        async with ClientSession(**kwargs) as client_session:
            yield client_session



from __future__ import annotations

import contextvars as cv
import json
from collections.abc import (
    Collection,
    Mapping,
    Sequence,
)
from contextlib import AbstractAsyncContextManager
from datetime import timedelta
from functools import cached_property, lru_cache
from itertools import chain
from typing import TypeVar

import sqlalchemy as sa
from loguru import logger
from typing_extensions import Self, TypeAlias

from . import http, pkg_db
from ._sources.cfcore import CfCoreResolver
from ._sources.github import GithubResolver
from ._sources.instawow import InstawowResolver
from ._sources.tukui import TukuiResolver
from ._sources.wago import WagoResolver
from ._sources.wowi import WowiResolver
from .catalogue.cataloguer import (
    CATALOGUE_VERSION,
    ComputedCatalogue,
)
from .config import Config
from .http import make_generic_progress_ctx
from .plugins import load_plugins
from .resolvers import Resolver
from .utils import (
    WeakValueDefaultDictionary,
    time_op,
)

_LOAD_CATALOGUE_LOCK = '_LOAD_CATALOGUE_'


_T = TypeVar('_T')


class _DummyLock:
    async def __aenter__(self):
        pass

    async def __aexit__(self, *args: object):
        pass


class _Resolvers(dict[str, Resolver]):
    @cached_property
    def priority_dict(self) -> _ResolverPriorityDict:
        return _ResolverPriorityDict(self)

    @cached_property
    def addon_toc_key_and_id_pairs(self) -> Collection[tuple[str, str]]:
        return [
            (r.metadata.addon_toc_key, r.metadata.id)
            for r in self.values()
            if r.metadata.addon_toc_key
        ]


class _ResolverPriorityDict(dict[str, float]):
    def __init__(self, resolvers: _Resolvers) -> None:
        super().__init__((n, i) for i, n in enumerate(resolvers))

    def __missing__(self, key: str) -> float:
        return float('inf')


_web_client = cv.ContextVar[http.ClientSessionType]('_web_client')

LocksType: TypeAlias = Mapping[object, AbstractAsyncContextManager[None]]

_locks = cv.ContextVar[LocksType]('_locks', default=WeakValueDefaultDictionary(_DummyLock))


def contextualise(
    *,
    web_client: http.ClientSessionType | None = None,
    locks: LocksType | None = None,
) -> None:
    "Set variables for the current context."
    if web_client is not None:
        _web_client.set(web_client)
    if locks is not None:
        _locks.set(locks)


@lru_cache(1)
def _parse_catalogue(raw_catalogue: bytes):
    with time_op(lambda t: logger.debug(f'parsed catalogue in {t:.3f}s')):
        return ComputedCatalogue.from_base_catalogue(json.loads(raw_catalogue))


class ManagerCtx:
    __slots__ = [
        'config',
        'database',
        'resolvers',
    ]

    RESOLVERS: Sequence[type[Resolver]] = [
        GithubResolver,
        CfCoreResolver,
        WowiResolver,
        TukuiResolver,
        InstawowResolver,
        WagoResolver,
    ]
    'Default resolvers.'

    _base_catalogue_url = (
        f'https://raw.githubusercontent.com/layday/instawow-data/data/'
        f'base-catalogue-v{CATALOGUE_VERSION}.compact.json'
    )
    _catalogue_ttl = timedelta(hours=4)

    def __init__(
        self,
        config: Config,
        database: sa.Engine,
    ) -> None:
        self.config: Config = config
        self.database: sa.Engine = database

        builtin_resolver_classes = list(self.RESOLVERS)

        for resolver, access_token in (
            (r, getattr(self.config.global_config.access_tokens, r.requires_access_token, None))
            for r in self.RESOLVERS
            if r.requires_access_token is not None
        ):
            if access_token is None:
                builtin_resolver_classes.remove(resolver)

        plugin_hook = load_plugins()
        resolver_classes = chain(
            (r for g in plugin_hook.instawow_add_resolvers() for r in g), builtin_resolver_classes
        )
        self.resolvers = _Resolvers((r.metadata.id, r(self)) for r in resolver_classes)

    @classmethod
    def from_config(cls, config: Config) -> Self:
        "Instantiate the manager from a configuration object."
        return cls(config, pkg_db.prepare_database(config.db_uri))

    @property
    def locks(self) -> LocksType:
        "Lock factory used to synchronise async operations."
        return _locks.get()

    @property
    def web_client(self) -> http.ClientSessionType:
        return _web_client.get()

    async def synchronise(self) -> ComputedCatalogue:
        "Fetch the catalogue from the interwebs and load it."
        async with self.locks[_LOAD_CATALOGUE_LOCK], self.web_client.get(
            self._base_catalogue_url,
            expire_after=self._catalogue_ttl,
            raise_for_status=True,
            trace_request_ctx=make_generic_progress_ctx('Synchronising catalogue'),
        ) as response:
            raw_catalogue = await response.read()
        return _parse_catalogue(raw_catalogue)



from __future__ import annotations

import os
import sqlite3
from collections.abc import Set
from contextlib import contextmanager
from typing import Any, NoReturn

from aiohttp_client_cache import BaseCache, CacheBackend, ResponseOrKey
from typing_extensions import LiteralString


@contextmanager
def acquire_cache_connection(parent_folder_path: os.PathLike[str]):
    with sqlite3.connect(os.path.join(parent_folder_path, '_aiohttp-cache.sqlite')) as connection:
        connection.execute('PRAGMA journal_mode = wal')
        connection.execute('PRAGMA synchronous = normal')
        yield connection


class SQLiteBackend(CacheBackend):
    def __init__(self, connection: sqlite3.Connection, **kwargs: Any):
        super().__init__(**kwargs)
        self.responses = _SQLitePickleCache().attach('responses', connection)
        self.redirects = _SQLiteSimpleCache().attach('redirects', connection)


class _SQLiteSimpleCache(BaseCache):
    def attach(self, table_name: LiteralString, connection: sqlite3.Connection):
        self.table_name = table_name
        self._connection = connection
        self._connection.execute(
            f'CREATE TABLE IF NOT EXISTS "{self.table_name}" (key PRIMARY KEY, value)'
        )
        return self

    async def clear(self) -> NoReturn:
        raise NotImplementedError

    async def contains(self, key: str):
        cursor = self._connection.execute(
            f'SELECT 1 FROM "{self.table_name}" WHERE key = ?',
            (key,),
        )
        (value,) = cursor.fetchone() or (0,)
        return value

    async def delete(self, key: str):
        self._connection.execute(
            f'DELETE FROM "{self.table_name}" WHERE key = ?',
            (key,),
        )
        self._connection.commit()

    async def bulk_delete(self, keys: Set[str]) -> NoReturn:
        raise NotImplementedError

    async def keys(self):
        cursor = self._connection.execute(f'SELECT key FROM "{self.table_name}"')
        for (value,) in cursor:
            yield value

    async def read(self, key: str):
        cursor = self._connection.execute(
            f'SELECT value FROM "{self.table_name}" WHERE key = ?',
            (key,),
        )
        (value,) = cursor.fetchone() or (None,)
        return value

    async def size(self):
        cursor = self._connection.execute(f'SELECT COUNT(key) FROM "{self.table_name}"')
        (value,) = cursor.fetchone()
        return value

    async def values(self):
        cursor = self._connection.execute(f'SELECT value FROM "{self.table_name}"')
        for (value,) in cursor:
            yield value

    async def write(self, key: str, item: ResponseOrKey):
        self._connection.execute(
            f'INSERT OR REPLACE INTO "{self.table_name}" (key, value) VALUES (?, ?)',
            (key, item),
        )
        self._connection.commit()


class _SQLitePickleCache(_SQLiteSimpleCache):
    async def read(self, key: str):
        return self.deserialize(await super().read(key))

    async def values(self):
        cursor = self._connection.execute(f'SELECT value FROM "{self.table_name}"')
        for (value,) in cursor:
            yield self.deserialize(value)

    async def write(self, key: str, item: ResponseOrKey):
        encoded_item = self.serialize(item)
        if encoded_item:
            await super().write(key, memoryview(encoded_item))



from __future__ import annotations

import pytest
from aresponses import ResponsesMockServer
from attrs import evolve

from instawow._version import is_outdated
from instawow.config import GlobalConfig


@pytest.mark.parametrize(
    '_iw_mock_aiohttp_requests',
    [set()],
    indirect=True,
)
async def test_is_outdated_works_in_variety_of_scenarios(
    monkeypatch: pytest.MonkeyPatch,
    aresponses: ResponsesMockServer,
    iw_global_config_values: dict[str, object],
):
    global_config = GlobalConfig.from_values(iw_global_config_values).write()

    # version == '0+dev', version not cached
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0+dev')
        assert await is_outdated(global_config) == (False, '')

    # Update check disabled, version not cached
    assert await is_outdated(evolve(global_config, auto_update_check=False)) == (False, '')

    # Endpoint not responsive, version not cached
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0.1.0')
        aresponses.add(
            'pypi.org',
            '/simple/instawow',
            'get',
            aresponses.Response(status=500),
        )
        assert await is_outdated(global_config) == (False, '0.1.0')

    # Endpoint responsive, version not cached and version different
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0.1.0')
        aresponses.add(
            'pypi.org',
            '/simple/instawow',
            'get',
            {'versions': ['1.0.0']},
        )
        assert await is_outdated(global_config) == (True, '1.0.0')

    # version == '0+dev', version cached
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0+dev')
        assert await is_outdated(global_config) == (False, '')

    # Update check disabled, version cached
    assert await is_outdated(evolve(global_config, auto_update_check=False)) == (False, '')

    # Endpoint not responsive, version cached
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0.1.0')
        aresponses.add(
            'pypi.org',
            '/simple/instawow',
            'get',
            aresponses.Response(status=500),
        )
        assert await is_outdated(global_config) == (True, '1.0.0')

    # Endpoint responsive, version cached and version same
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0.1.0')
        aresponses.add(
            'pypi.org',
            '/simple/instawow',
            'get',
            {'versions': ['1.0.0']},
        )
        assert await is_outdated(global_config) == (True, '1.0.0')

    # Endpoint responsive, version cached and version different
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '1.0.0')
        aresponses.add(
            'pypi.org',
            '/simple/instawow',
            'get',
            {'versions': ['1.0.0']},
        )
        assert await is_outdated(global_config) == (False, '1.0.0')



from __future__ import annotations

from datetime import timedelta

from typing_extensions import TypedDict

from . import config

__version__ = '3.1.0'


class _SimpleApiProject(TypedDict):
    versions: list[str]


async def is_outdated(global_config: config.GlobalConfig) -> tuple[bool, str]:
    """Check on PyPI to see if instawow is outdated.

    The response is cached for 24 hours.
    """

    from aiohttp import ClientError
    from packaging.version import Version

    from . import __version__
    from .http import init_web_client

    if not global_config.auto_update_check:
        return (False, '')

    parsed_version = Version(__version__)
    if parsed_version.local:
        return (False, '')

    try:
        async with init_web_client(
            global_config.cache_dir, raise_for_status=True
        ) as web_client, web_client.get(
            'https://pypi.org/simple/instawow',
            expire_after=timedelta(days=1),
            headers={
                'Accept': 'application/vnd.pypi.simple.v1+json',
            },
        ) as response:
            metadata: _SimpleApiProject = await response.json()
            version = max(metadata['versions'], key=Version)
    except ClientError:
        version = __version__

    return (Version(version) > parsed_version, version)



from __future__ import annotations

import asyncio
import enum
import sys
import textwrap
from collections.abc import Awaitable, Callable, Collection, Iterable, Mapping, Sequence
from datetime import datetime, timezone
from functools import cached_property, partial
from itertools import chain, repeat
from pathlib import Path
from typing import Any, Generic, NoReturn, TypeVar, overload

import click
import click.types
from attrs import asdict, evolve, fields, resolve_types
from loguru import logger
from typing_extensions import Self

from . import __version__, manager_ctx, pkg_db, pkg_management, pkg_models
from . import results as R
from ._logging import setup_logging
from .common import ChangelogFormat, Defn, Flavour, SourceMetadata, Strategy
from .config import Config, GlobalConfig, config_converter
from .http import TraceRequestCtx, init_web_client
from .plugins import load_plugins
from .utils import StrEnum, all_eq, gather, reveal_folder, tabulate, uniq

_T = TypeVar('_T')
_TStrEnum = TypeVar('_TStrEnum', bound=StrEnum)


def _patch_asyncio():
    # See https://github.com/aio-libs/aiohttp/issues/4324.
    # We are not using pipes on Windows or async subprocesses so it's ok
    # to just override the loop policy.
    if sys.platform == 'win32' and sys.version_info < (3, 10):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


@logger.catch(reraise=True)
def main(*args: Any, **kwargs: Any) -> None:
    _patch_asyncio()
    cli(*args, **kwargs)


class Report:
    SUCCESS_SYMBOL = click.style('✓', fg='green')
    FAILURE_SYMBOL = click.style('✗', fg='red')
    WARNING_SYMBOL = click.style('!', fg='blue')

    def __init__(
        self,
        results: Iterable[tuple[Defn, R.Result]],
        filter_fn: Callable[[R.Result], bool] = lambda _: True,
    ) -> None:
        self.results = list(results)
        self.filter_fn = filter_fn

    @property
    def exit_code(self) -> int:
        return any(
            isinstance(r, (R.ManagerError, R.InternalError)) and self.filter_fn(r)
            for _, r in self.results
        )

    @classmethod
    def _result_type_to_symbol(cls, result: R.Result) -> str:
        if isinstance(result, R.InternalError):
            return cls.WARNING_SYMBOL
        elif isinstance(result, R.ManagerError):
            return cls.FAILURE_SYMBOL
        else:
            return cls.SUCCESS_SYMBOL

    def __str__(self) -> str:
        return '\n'.join(
            f'{self._result_type_to_symbol(r)} {click.style(a.as_uri(), bold=True)}\n'
            f'{textwrap.fill(r.message, initial_indent=" " * 2, subsequent_indent=" " * 4)}'
            for a, r in self.results
            if self.filter_fn(r)
        )

    def generate(self) -> None:
        mw: _CtxObjWrapper | None = click.get_current_context().obj

        if mw and mw.manager.ctx.config.global_config.auto_update_check:
            from ._version import is_outdated

            outdated, new_version = mw.run_with_progress(
                is_outdated(mw.manager.ctx.config.global_config)
            )
            if outdated:
                click.echo(f'{self.WARNING_SYMBOL} instawow v{new_version} is available')

        report = str(self)
        if report:
            click.echo(report)

    def generate_and_exit(self) -> NoReturn:
        self.generate()
        ctx = click.get_current_context()
        ctx.exit(self.exit_code)


class _CtxObjWrapper:
    def __init__(self, ctx: click.Context) -> None:
        self._ctx = ctx

    @cached_property
    def manager(self) -> pkg_management.PkgManager:
        global_config = GlobalConfig.read().ensure_dirs()
        try:
            config = Config.read(global_config, self._ctx.params['profile']).ensure_dirs()
        except FileNotFoundError:
            config = self._ctx.invoke(configure)

        setup_logging(config.logging_dir, *self._ctx.params['debug'])

        ctx = manager_ctx.ManagerCtx.from_config(config)
        return pkg_management.PkgManager(ctx)

    def run_with_progress(self, awaitable: Awaitable[_T]) -> _T:
        cache_dir = self.manager.ctx.config.global_config.cache_dir
        params = self._ctx.params

        if any(params['debug']):

            async def run():
                async with init_web_client(cache_dir, no_cache=params['no_cache']) as web_client:
                    manager_ctx.contextualise(web_client=web_client)
                    return await awaitable

        else:
            from contextlib import contextmanager

            import aiohttp
            from prompt_toolkit.shortcuts import ProgressBar, ProgressBarCounter

            from ._cli_prompts import make_progress_bar

            def init_cli_web_client(progress_bar: ProgressBar, tickers: set[asyncio.Task[None]]):
                TICK_INTERVAL = 0.1

                async def do_on_request_end(
                    client_session: aiohttp.ClientSession,
                    trace_config_ctx: Any,
                    params: aiohttp.TraceRequestEndParams,
                ):
                    trace_request_ctx: TraceRequestCtx = trace_config_ctx.trace_request_ctx
                    if trace_request_ctx:
                        response = params.response
                        label = (
                            'Downloading '
                            + Defn(
                                trace_request_ctx['defn'].source, trace_request_ctx['defn'].alias
                            ).as_uri()
                            if trace_request_ctx['report_progress'] == 'pkg_download'
                            else trace_request_ctx['label']
                        )
                        # When the total is ``None`` the progress bar is
                        # in an "indeterminate" state.
                        # We cannot display progress for encoded responses because
                        # the size before decoding is not exposed by the
                        # aiohttp streaming API
                        total = (
                            None
                            if aiohttp.hdrs.CONTENT_ENCODING in response.headers
                            else response.content_length
                        )

                        counters = progress_bar.counters

                        async def ticker():
                            counter = ProgressBarCounter[object](
                                progress_bar=progress_bar, label=label, total=total
                            )
                            counters.append(counter)
                            try:
                                while not response.content.is_eof():
                                    counter.items_completed = response.content.total_bytes
                                    progress_bar.invalidate()
                                    await asyncio.sleep(TICK_INTERVAL)
                            finally:
                                counters.remove(counter)

                        tickers.add(asyncio.create_task(ticker()))

                trace_config = aiohttp.TraceConfig()
                trace_config.on_request_end.append(do_on_request_end)
                trace_config.freeze()
                return init_web_client(
                    cache_dir, no_cache=params['no_cache'], trace_configs=[trace_config]
                )

            @contextmanager
            def cancel_tickers(progress_bar: ProgressBar):
                tickers: set[asyncio.Task[None]] = set()
                try:
                    yield tickers
                finally:
                    for ticker in tickers:
                        ticker.cancel()
                    progress_bar.invalidate()

            async def run():
                with make_progress_bar() as progress_bar, cancel_tickers(progress_bar) as tickers:
                    async with init_cli_web_client(progress_bar, tickers) as web_client:
                        manager_ctx.contextualise(web_client=web_client)
                        return await awaitable

        return asyncio.run(run())


def _with_manager(fn: Callable[..., object]):
    def wrapper(ctx: click.Context, __: click.Parameter, value: object):
        return fn(ctx.obj.manager, value)

    return wrapper


class _StrEnumChoiceParam(Generic[_TStrEnum], click.Choice):
    def __init__(
        self,
        choice_enum: type[_TStrEnum],
        case_sensitive: bool = True,
    ) -> None:
        super().__init__(
            choices=list(choice_enum),
            case_sensitive=case_sensitive,
        )
        self.__choice_enum = choice_enum

    def convert(
        self, value: Any, param: click.Parameter | None, ctx: click.Context | None
    ) -> _TStrEnum:
        converted_value = super().convert(value, param, ctx)
        return self.__choice_enum(converted_value)


class _ManyOptionalChoiceValueParam(click.types.CompositeParamType):
    name = 'optional-choice-value'

    def __init__(
        self,
        choice_param: click.Choice,
        *,
        value_types: Mapping[str, click.types.ParamType] = {},
    ) -> None:
        super().__init__()
        self.__choice_param = choice_param
        self.__value_types = value_types

    def __parse_value(self, value: tuple[str, ...]):
        return (
            (k, v, self.__choice_param, vc)
            for r in value
            for k, _, v in (r.partition('='),)
            for vc in (self.__value_types.get(k),)
        )

    @property
    def arity(self) -> int:
        return -1

    def convert(
        self, value: tuple[str, ...], param: click.Parameter | None, ctx: click.Context | None
    ) -> Any:
        return {
            kc.convert(k, param, ctx): vc.convert(v, param, ctx) if vc and v else (v or None)
            for k, v, kc, vc in self.__parse_value(value)
        }

    def get_metavar(self, param: click.Parameter) -> str:
        return f'{{{",".join(self.__choice_param.choices)}}}[=VALUE]'


def _register_plugin_commands(group: click.Group):
    plugin_hook = load_plugins()
    additional_commands = (c for g in plugin_hook.instawow_add_commands() for c in g)
    for command in additional_commands:
        group.add_command(command)
    return group


def _parse_debug_option(
    _: click.Context, __: click.Parameter, value: float
) -> tuple[bool, bool, bool]:
    return (value > 0, value > 1, value > 2)


@_register_plugin_commands
@click.group(context_settings={'help_option_names': ('-h', '--help')})
@click.version_option(__version__, prog_name=__spec__.parent)
@click.option(
    '--debug',
    '-d',
    count=True,
    help='Log incrementally more things.  Additive.',
    callback=_parse_debug_option,
)
@click.option(
    '--no-cache',
    is_flag=True,
    default=False,
    help='Disable the HTTP cache.',
)
@click.option(
    '--profile',
    '-p',
    default='__default__',
    help='Activate the specified profile.',
)
@click.pass_context
def cli(ctx: click.Context, **__: object) -> None:
    "Add-on manager for World of Warcraft."
    ctx.obj = _CtxObjWrapper(ctx)


@overload
def _parse_uri(
    manager: pkg_management.PkgManager,
    value: str,
    *,
    raise_invalid: bool = True,
    include_strategies: bool = False,
) -> Defn:
    ...


@overload
def _parse_uri(
    manager: pkg_management.PkgManager,
    value: list[str],
    *,
    raise_invalid: bool = True,
    include_strategies: bool = False,
) -> list[Defn]:
    ...


def _parse_uri(
    manager: pkg_management.PkgManager,
    value: str | list[str] | None,
    *,
    raise_invalid: bool = True,
    include_strategies: bool = False,
) -> Defn | list[Defn] | None:
    if value is None:
        return None

    if not isinstance(value, str):
        defns = (
            _parse_uri(
                manager, v, raise_invalid=raise_invalid, include_strategies=include_strategies
            )
            for v in value
        )
        return uniq(defns)

    try:
        defn = Defn.from_uri(
            value,
            known_sources=manager.ctx.resolvers,
            allow_unsourced=True,
            include_strategies=include_strategies,
        )
    except ValueError as exc:
        raise click.BadParameter(exc.args[0]) from None

    if defn.is_unsourced:
        match = manager.pair_uri(defn.alias)
        if match:
            source, alias = match
            defn = evolve(defn, source=source, alias=alias)
        elif raise_invalid and ':' not in defn.alias:
            raise click.BadParameter(value)

    return defn


@cli.command
@click.argument(
    'addons', nargs=-1, callback=_with_manager(partial(_parse_uri, include_strategies=True))
)
@click.option(
    '--replace',
    is_flag=True,
    default=False,
    help='Replace unreconciled add-ons.',
)
@click.option(
    '--dry-run',
    is_flag=True,
    default=False,
    help='Pretend to install add-ons.  Add-on archives will not be download and the '
    'database will not be modified.',
)
@click.pass_obj
def install(
    mw: _CtxObjWrapper,
    addons: Sequence[Defn],
    replace: bool,
    dry_run: bool,
) -> None:
    "Install add-ons."
    results = mw.run_with_progress(mw.manager.install(addons, replace, dry_run))
    Report(results.items()).generate_and_exit()


@cli.command
@click.argument(
    'addons', nargs=-1, callback=_with_manager(partial(_parse_uri, include_strategies=True))
)
@click.option(
    '--retain-strategies',
    is_flag=True,
    default=False,
    help='Respect the strategies of [ADDONS] if they result in a change.',
)
@click.option(
    '--dry-run',
    is_flag=True,
    default=False,
    help='Pretend to update add-ons.  Add-on archives will not be download and the '
    'database will not be modified.  Use this option to check for updates.',
)
@click.pass_obj
def update(
    mw: _CtxObjWrapper,
    addons: Sequence[Defn],
    retain_strategies: bool,
    dry_run: bool,
) -> None:
    "Update installed add-ons."
    import sqlalchemy as sa

    def filter_results(result: R.Result):
        # Hide packages from output if they are up to date
        # and ``update`` was invoked without args,
        # provided that they are not pinned
        if addons or not isinstance(result, R.PkgUpToDate):
            return True
        else:
            return result.is_pinned

    def installed_pkgs_to_defns():
        with mw.manager.ctx.database.connect() as connection:
            return [
                pkg_models.Pkg.from_row_mapping(connection, p).to_defn()
                for p in connection.execute(sa.select(pkg_db.pkg)).mappings().all()
            ]

    results = mw.run_with_progress(
        mw.manager.update(
            addons or installed_pkgs_to_defns(), retain_strategies if addons else False, dry_run
        )
    )
    Report(results.items(), filter_results).generate_and_exit()


@cli.command
@click.argument('addons', nargs=-1, required=True, callback=_with_manager(_parse_uri))
@click.option(
    '--keep-folders',
    is_flag=True,
    default=False,
    help='Remove the add-on from the database but do not delete its folders.',
)
@click.pass_obj
def remove(mw: _CtxObjWrapper, addons: Sequence[Defn], keep_folders: bool) -> None:
    "Remove add-ons."
    results = mw.run_with_progress(mw.manager.remove(addons, keep_folders))
    Report(results.items()).generate_and_exit()


@cli.command
@click.argument('addon', callback=_with_manager(_parse_uri))
@click.option(
    '--undo',
    is_flag=True,
    default=False,
    help='Undo rollback by reinstalling an add-on using the default strategy.',
)
@click.pass_obj
def rollback(mw: _CtxObjWrapper, addon: Defn, undo: bool) -> None:
    "Roll an add-on back to an older version."
    from ._cli_prompts import Choice, ask, select

    pkg = mw.manager.get_pkg(addon)
    if not pkg:
        Report([(addon, R.PkgNotInstalled())]).generate_and_exit()
    elif Strategy.VersionEq not in mw.manager.ctx.resolvers[pkg.source].metadata.strategies:
        Report([(addon, R.PkgStrategiesUnsupported({Strategy.VersionEq}))]).generate_and_exit()
    elif undo:
        Report(mw.run_with_progress(mw.manager.update([addon], True)).items()).generate_and_exit()

    reconstructed_defn = pkg.to_defn()

    versions = pkg.logged_versions
    if len(versions) <= 1:
        Report([(addon, R.PkgFilesMissing('cannot find older versions'))]).generate_and_exit()

    choices = [
        Choice(
            [('', v.version)],
            value=v.version,
            disabled='installed version' if v.version == pkg.version else None,
        )
        for v in versions
    ]
    selection = ask(
        select(f'Select version of {reconstructed_defn.as_uri()} for rollback', choices)
    )

    Report(
        mw.run_with_progress(
            mw.manager.update([reconstructed_defn.with_version(selection)], True)
        ).items()
    ).generate_and_exit()


@cli.command
@click.option('--auto', '-a', is_flag=True, default=False, help='Do not ask for user input.')
@click.option(
    '--installed',
    'rereconcile',
    is_flag=True,
    default=False,
    help='Re-reconcile installed add-ons.',
)
@click.option(
    '--list-unreconciled', is_flag=True, default=False, help='List unreconciled add-ons and exit.'
)
@click.pass_obj
def reconcile(mw: _CtxObjWrapper, auto: bool, rereconcile: bool, list_unreconciled: bool) -> None:
    "Reconcile pre-installed add-ons."
    from ._cli_prompts import PkgChoice, ask, confirm, select, skip
    from .matchers import DEFAULT_MATCHERS, AddonFolder, get_unreconciled_folders

    def construct_choice(pkg: pkg_models.Pkg, highlight_version: bool, disabled: bool):
        defn = pkg.to_defn()
        return PkgChoice(
            [
                ('', f'{defn.as_uri()}=='),
                ('class:highlight-sub' if highlight_version else '', pkg.version),
            ],
            pkg=pkg,
            value=defn,
            disabled=disabled,
        )

    def gather_selections(
        groups: Collection[tuple[_T, Sequence[Defn]]],
        selector: Callable[[_T, Sequence[pkg_models.Pkg]], Defn | None],
    ):
        results = mw.run_with_progress(mw.manager.resolve(uniq(d for _, b in groups for d in b)))
        for addons_or_pkg, defns in groups:
            shortlist = [r for d in defns for r in (results[d],) if isinstance(r, pkg_models.Pkg)]
            if shortlist:
                selection = selector(addons_or_pkg, shortlist)
                yield selection
            else:
                yield None

    if rereconcile:
        if auto:
            raise click.UsageError('Cannot use "--auto" with "--installed"')

        import sqlalchemy as sa

        def select_alternative_pkg(installed_pkg: pkg_models.Pkg, pkgs: Sequence[pkg_models.Pkg]):
            highlight_version = not all_eq(i.version for i in (installed_pkg, *pkgs))
            choices = [
                construct_choice(installed_pkg, highlight_version, True),
                *(construct_choice(p, highlight_version, False) for p in pkgs),
                skip,
            ]
            selection = ask(select(installed_pkg.name, choices))
            return selection or None

        with mw.manager.ctx.database.connect() as connection:
            installed_pkgs = [
                pkg_models.Pkg.from_row_mapping(connection, p)
                for p in connection.execute(
                    sa.select(pkg_db.pkg).order_by(sa.func.lower(pkg_db.pkg.c.name))
                )
                .mappings()
                .all()
            ]

        groups = mw.run_with_progress(mw.manager.find_equivalent_pkg_defns(installed_pkgs))
        selections = [
            (p, s)
            for (p, _), s in zip(
                groups.items(), gather_selections(groups.items(), select_alternative_pkg)
            )
            if s
        ]
        if selections and ask(confirm('Install selected add-ons?')):
            Report(
                mw.run_with_progress(
                    mw.manager.remove([p.to_defn() for p, _ in selections], False),
                ).items()
            ).generate()
            Report(
                mw.run_with_progress(
                    mw.manager.install([s for _, s in selections], False),
                ).items()
            ).generate_and_exit()

    else:
        PREAMBLE = textwrap.dedent(
            '''\
            Use the arrow keys to navigate, <o> to open an add-on in your browser,
            enter to make a selection and <s> to skip to the next item.

            Versions that differ from the installed version or differ between
            choices are highlighted in purple.

            instawow will perform three passes in decreasing order of accuracy,
            looking to match source IDs and add-on names in TOC files, and folders.

            Selected add-ons will be reinstalled.

            You can also run `reconcile` in automatic mode by passing
            the `--auto` flag.  In this mode, add-ons will be reconciled
            without user input.
            '''
        )

        leftovers = get_unreconciled_folders(mw.manager.ctx)
        if list_unreconciled:
            table_rows = [('unreconciled',), *((f.name,) for f in sorted(leftovers))]
            click.echo(tabulate(table_rows))
            return
        elif not leftovers:
            click.echo('No add-ons left to reconcile.')
            return

        if not auto:
            click.echo(PREAMBLE)

        def select_pkg(addons: Sequence[AddonFolder], pkgs: Sequence[pkg_models.Pkg]):
            def combine_names():
                return textwrap.shorten(', '.join(a.name for a in addons), 60)

            # Highlight version if there's multiple of them
            highlight_version = not all_eq(i.version for i in chain(addons, pkgs))
            choices = [
                *(construct_choice(p, highlight_version, False) for p in pkgs),
                skip,
            ]
            selection = ask(select(f'{combine_names()} [{addons[0].version or "?"}]', choices))
            return selection or None

        def pick_first_pkg(addons: Sequence[AddonFolder], pkgs: Sequence[pkg_models.Pkg]):
            return pkgs[0].to_defn()

        select_pkg_ = pick_first_pkg if auto else select_pkg
        confirm_install = (
            (lambda: True) if auto else (lambda: ask(confirm('Install selected add-ons?')))
        )

        for fn in DEFAULT_MATCHERS.values():
            groups = mw.run_with_progress(fn(mw.manager.ctx, leftovers))
            selections = [s for s in gather_selections(groups, select_pkg_) if s is not None]
            if selections and confirm_install():
                results = mw.run_with_progress(mw.manager.install(selections, True))
                Report(results.items()).generate()

            leftovers = get_unreconciled_folders(mw.manager.ctx)
            if not leftovers:
                break

        if leftovers:
            click.echo()
            table_rows = [('unreconciled',), *((f.name,) for f in sorted(leftovers))]
            click.echo(tabulate(table_rows))


def _concat_search_terms(_: click.Context, __: click.Parameter, value: tuple[str, ...]):
    return ' '.join(value)


def _parse_iso_date_into_datetime(_: click.Context, __: click.Parameter, value: str | None):
    if value is not None:
        return datetime.strptime(value, '%Y-%m-%d').replace(tzinfo=timezone.utc)


@cli.command
@click.argument('search-terms', nargs=-1, required=True, callback=_concat_search_terms)
@click.option(
    '--limit',
    '-l',
    default=10,
    show_default=True,
    type=click.IntRange(1, 20, clamp=True),
    help='Maximum results to return.',
)
@click.option(
    '--start-date',
    callback=_parse_iso_date_into_datetime,
    help='Omit results before this date.',
    metavar='YYYY-MM-DD',
)
@click.option(
    '--source',
    'sources',
    multiple=True,
    help='Search only in the specified source.  Multiple option.',
)
@click.option(
    '--prefer-source',
    help='Hide duplicates from other sources.',
)
@click.option(
    '--no-exclude-installed',
    is_flag=True,
    default=False,
    help='Do not exclude installed add-ons from search results.',
)
@click.pass_context
def search(
    ctx: click.Context,
    search_terms: str,
    limit: int,
    sources: Sequence[str],
    prefer_source: str | None,
    start_date: datetime | None,
    no_exclude_installed: bool,
) -> None:
    "Search for add-ons to install."
    from ._cli_prompts import PkgChoice, ask, checkbox, confirm
    from .catalogue.search import search

    mw: _CtxObjWrapper = ctx.obj

    catalogue_entries = mw.run_with_progress(
        search(
            mw.manager.ctx,
            search_terms,
            limit=limit,
            sources=frozenset(sources),
            prefer_source=prefer_source,
            start_date=start_date,
            filter_installed='ident' if no_exclude_installed else 'exclude_from_all_sources',
        )
    )
    results = mw.run_with_progress(
        mw.manager.resolve([Defn(e.source, e.id) for e in catalogue_entries])
    )
    pkgs, _ = pkg_management.bucketise_results(results.items())
    if pkgs:
        choices = [
            PkgChoice(f'{p.name}  ({e.as_uri()}=={p.version})', e, pkg=p)
            for d, p in pkgs.items()
            for e in (evolve(d, alias=p.slug, id=p.id),)
        ]
        selections: list[Defn] = ask(checkbox('Select add-ons to install', choices=choices))
        if selections:
            if ask(confirm('Install selected add-ons?')):
                ctx.invoke(install, addons=selections, replace=False, dry_run=False)
        else:
            click.echo(
                'Nothing was selected; select add-ons with <space>'
                ' and confirm by pressing <enter>.'
            )

    else:
        click.echo('No results found.')


class _ListFormat(StrEnum):
    Simple = enum.auto()
    Detailed = enum.auto()
    Json = enum.auto()


@cli.command('list')
@click.argument(
    'addons', nargs=-1, callback=_with_manager(partial(_parse_uri, raise_invalid=False))
)
@click.option(
    '--format',
    '-f',
    'output_format',
    type=_StrEnumChoiceParam(_ListFormat),
    default=_ListFormat.Simple,
    show_default=True,
    help='Change the output format.',
)
@click.pass_obj
def list_installed(mw: _CtxObjWrapper, addons: Sequence[Defn], output_format: _ListFormat) -> None:
    "List installed add-ons."
    import sqlalchemy as sa

    with mw.manager.ctx.database.connect() as connection:

        def format_deps(pkg: pkg_models.Pkg):
            return (
                Defn(pkg.source, s or e.id).as_uri()
                for e in pkg.deps
                for s in (
                    connection.execute(
                        sa.select(pkg_db.pkg.c.slug).filter_by(source=pkg.source, id=e.id)
                    ).scalar_one_or_none(),
                )
            )

        def row_mappings_to_pkgs():
            return map(pkg_models.Pkg.from_row_mapping, repeat(connection), pkg_mappings)

        pkg_select_query = sa.select(pkg_db.pkg)
        if addons:
            pkg_select_query = pkg_select_query.filter(
                sa.or_(
                    *(
                        pkg_db.pkg.c.slug.contains(d.alias)
                        if d.is_unsourced
                        else (pkg_db.pkg.c.source == d.source)
                        & (
                            (pkg_db.pkg.c.id == d.alias)
                            | (sa.func.lower(pkg_db.pkg.c.slug) == sa.func.lower(d.alias))
                        )
                        for d in addons
                    )
                )
            )
        pkg_mappings = (
            connection.execute(
                pkg_select_query.order_by(pkg_db.pkg.c.source, sa.func.lower(pkg_db.pkg.c.name))
            )
            .mappings()
            .all()
        )

        if output_format is _ListFormat.Json:
            from cattrs.preconf.json import make_converter

            json_converter = make_converter()
            click.echo(
                json_converter.dumps(
                    row_mappings_to_pkgs(),
                    list[pkg_models.Pkg],
                    indent=2,
                )
            )

        elif output_format is _ListFormat.Detailed:
            formatter = click.HelpFormatter(max_width=99)
            for pkg in row_mappings_to_pkgs():
                with formatter.section(pkg.to_defn().as_uri()):
                    formatter.write_dl(
                        [
                            ('name', pkg.name),
                            ('description', textwrap.shorten(pkg.description, 280)),
                            ('url', pkg.url),
                            ('version', pkg.version),
                            ('date_published', pkg.date_published.isoformat(' ', 'minutes')[:-6]),
                            ('folders', ', '.join(f.name for f in pkg.folders)),
                            ('deps', ', '.join(format_deps(pkg))),
                            (
                                'options',
                                '\n'.join(f'{s}={v}' for s, v in asdict(pkg.options).items()),
                            ),
                        ]
                    )
            click.echo(formatter.getvalue(), nl=False)

        else:
            click.echo(
                ''.join(f'{Defn(p["source"], p["slug"]).as_uri()}\n' for p in pkg_mappings),
                nl=False,
            )


@cli.command(hidden=True)
@click.argument('addon', callback=_with_manager(partial(_parse_uri, raise_invalid=False)))
@click.pass_context
def info(ctx: click.Context, addon: Defn) -> None:
    "Alias of `list -f detailed`."
    ctx.invoke(list_installed, addons=(addon,), output_format=_ListFormat.Detailed)


@cli.command
@click.argument('addon', callback=_with_manager(partial(_parse_uri, raise_invalid=False)))
@click.pass_obj
def reveal(mw: _CtxObjWrapper, addon: Defn) -> None:
    "Bring an add-on up in your file manager."
    pkg = mw.manager.get_pkg(addon, partial_match=True)
    if pkg:
        reveal_folder(mw.manager.ctx.config.addon_dir / pkg.folders[0].name)
    else:
        Report([(addon, R.PkgNotInstalled())]).generate_and_exit()


@cli.command
@click.argument(
    'addon', callback=_with_manager(partial(_parse_uri, raise_invalid=False)), required=False
)
@click.option(
    '--convert/--no-convert',
    default=True,
    show_default=True,
    help='Convert HTML and Markdown changelogs to plain text using pandoc.',
)
@click.pass_obj
def view_changelog(mw: _CtxObjWrapper, addon: Defn | None, convert: bool) -> None:
    """View the changelog of an installed add-on.

    If `addon` is not provided, displays the changelogs of all add-ons
    to have been installed within one minute of the last add-on.
    """

    MAX_LINES = 100

    def make_converter():
        import shutil

        pandoc = shutil.which('pandoc')
        if pandoc is None:

            def noop_convert(source: str, changelog: str):
                return changelog

            return noop_convert
        else:
            import subprocess

            def real_convert(source: str, changelog: str):
                changelog_format = mw.manager.ctx.resolvers[source].metadata.changelog_format
                if changelog_format not in {ChangelogFormat.Html, ChangelogFormat.Markdown}:
                    return changelog
                else:
                    INPUT_FORMAT_CORRESPONDENCES = {
                        ChangelogFormat.Html: 'html',
                        # The "markdown" format will treat a list without a preceding
                        # empty line as a paragraph, which breaks the changelog
                        # of at least one popular add-on.
                        ChangelogFormat.Markdown: 'commonmark',
                    }

                    return subprocess.check_output(
                        [
                            pandoc,
                            '-f',
                            INPUT_FORMAT_CORRESPONDENCES[changelog_format],
                            '-t',
                            'plain',
                        ],
                        input=changelog,
                        text=True,
                    )

            return real_convert

    def format_combined_changelog_entry(source: str, slug: str, changelog: str):
        lines = changelog.splitlines()
        body = '\n'.join(
            (
                *(f'  {i}' for i in lines[:MAX_LINES]),
                *(('  [...]',) if len(lines) > MAX_LINES else ()),
            )
        )
        return f'{Defn(source, slug).as_uri()}:\n{body}'

    if addon:
        pkg = mw.manager.get_pkg(addon, partial_match=True)
        if pkg:
            changelog = mw.run_with_progress(
                mw.manager.get_changelog(pkg.source, pkg.changelog_url)
            )
            if convert:
                changelog = make_converter()(pkg.source, changelog)

            click.echo_via_pager(format_combined_changelog_entry(pkg.source, pkg.slug, changelog))

        else:
            Report([(addon, R.PkgNotInstalled())]).generate_and_exit()

    else:
        import sqlalchemy as sa

        with mw.manager.ctx.database.connect() as connection:
            join_clause = (pkg_db.pkg.c.source == pkg_db.pkg_version_log.c.pkg_source) & (
                pkg_db.pkg.c.id == pkg_db.pkg_version_log.c.pkg_id
            )
            last_installed_changelog_urls = connection.execute(
                sa.select(pkg_db.pkg.c.source, pkg_db.pkg.c.slug, pkg_db.pkg.c.changelog_url)
                .join(pkg_db.pkg_version_log, join_clause)
                .filter(
                    pkg_db.pkg_version_log.c.install_time
                    >= sa.select(
                        sa.func.datetime(
                            sa.func.max(pkg_db.pkg_version_log.c.install_time), '-1 minute'
                        )
                    )
                    .join(pkg_db.pkg, join_clause)
                    .scalar_subquery()
                )
            ).all()

        changelogs = mw.run_with_progress(
            gather(
                mw.manager.get_changelog(m.source, m.changelog_url)
                for m in last_installed_changelog_urls
            )
        )
        if convert:
            do_convert = make_converter()
            changelogs = (
                do_convert(m.source, c) for m, c in zip(last_installed_changelog_urls, changelogs)
            )

        click.echo_via_pager(
            '\n\n'.join(
                format_combined_changelog_entry(m.source, m.slug, c)
                for m, c in zip(last_installed_changelog_urls, changelogs)
            )
        )


@cli.command
@click.option(
    '--format',
    '-f',
    'output_format',
    type=click.Choice([_ListFormat.Json]),
    default=_ListFormat.Json,
    show_default=True,
    help='Change the output format.',
)
@click.pass_obj
def list_sources(mw: _CtxObjWrapper, output_format: _ListFormat) -> None:
    "Print source metadata."
    from cattrs.preconf.json import make_converter

    json_converter = make_converter()
    click.echo(
        json_converter.dumps(
            [r.metadata for r in mw.manager.ctx.resolvers.values()],
            list[SourceMetadata],
            indent=2,
        )
    )


def _show_active_config(ctx: click.Context, __: click.Parameter, value: bool):
    if value:
        click.echo(ctx.obj.manager.ctx.config.encode_for_display())
        ctx.exit()


async def _github_oauth_flow():
    from .github_auth import get_codes, poll_for_access_token

    async with init_web_client(None) as web_client:
        codes = await get_codes(web_client)
        click.echo(f'Navigate to {codes["verification_uri"]} and paste the code below:')
        click.echo(f'  {codes["user_code"]}')
        click.echo('Waiting...')
        access_token = await poll_for_access_token(
            web_client, codes['device_code'], codes['interval']
        )
        return access_token


class _EditableConfigOptions(StrEnum):
    AddonDir = 'addon_dir'
    GameFlavour = 'game_flavour'
    AutoUpdateCheck = 'global_config.auto_update_check'
    GithubAccessToken = 'global_config.access_tokens.github'
    CfcoreAccessToken = 'global_config.access_tokens.cfcore'
    WagoAddonsAccessToken = 'global_config.access_tokens.wago_addons'

    path: tuple[str, ...]

    def __new__(cls, value: str) -> Self:
        self = str.__new__(cls, value)
        self.path = tuple(value.split('.'))
        return self


@cli.command
@click.option(
    '--show-active',
    is_flag=True,
    default=False,
    expose_value=False,
    is_eager=True,
    callback=_show_active_config,
    help='Show the active configuration and exit.',
)
@click.argument(
    'editable-config-values',
    nargs=-1,
    type=_ManyOptionalChoiceValueParam(
        _StrEnumChoiceParam(_EditableConfigOptions),
        value_types={
            _EditableConfigOptions.AutoUpdateCheck: click.types.BoolParamType(),
        },
    ),
)
@click.pass_context
def configure(
    ctx: click.Context,
    editable_config_values: Mapping[_EditableConfigOptions, Any],
) -> Config:
    """Configure instawow.

    You can pass configuration keys as arguments to reconfigure an existing
    profile.  Pass a value to suppress the interactive prompt.  For example:

    \b
    * ``configure addon_dir`` will initiate an interactive session
      with autocompletion
    * ``configure "addon_dir=~/foo"` will set ``addon_dir``'s value
      to ``~/foo`` immediately
    """
    from ._cli_prompts import (
        AttrsFieldValidator,
        Choice,
        ask,
        confirm,
        password,
        path,
        select,
        skip,
    )
    from .wow_installations import (
        ADDON_DIR_PARTS,
        find_installations,
        infer_flavour_from_addon_dir,
    )

    profile = ctx.find_root().params['profile']

    existing_global_config = GlobalConfig.read()

    config_values: dict[str, Any] | None = None
    try:
        config_values = asdict(Config.read(existing_global_config, profile))
    except FileNotFoundError:
        pass
    except Exception:
        logger.exception('unable to read existing config')

    is_new_profile = config_values is None
    if is_new_profile:
        config_values = {'profile': profile, 'global_config': asdict(existing_global_config)}

    editable_config_values = dict(editable_config_values)
    if not editable_config_values:
        default_keys = {
            _EditableConfigOptions.AddonDir,
            _EditableConfigOptions.GameFlavour,
        }
        if existing_global_config.access_tokens.github is None:
            default_keys.add(_EditableConfigOptions.GithubAccessToken)

        editable_config_values = dict.fromkeys(default_keys)

    interactive_editable_config_keys = {k for k, v in editable_config_values.items() if v is None}

    # 0 = Unconfigured
    # 1 = `addon_dir` configured
    # 2 = both `addon_dir` and `game_flavour` configured
    installation_configured = 0
    if (
        is_new_profile
        and {_EditableConfigOptions.AddonDir, _EditableConfigOptions.GameFlavour}
        <= interactive_editable_config_keys
    ):
        known_intallations = {
            Config.read(existing_global_config, p).addon_dir.parent.parent
            for p in existing_global_config.list_profiles()
        }
        available_installations = [
            (k, v)
            for k, v in find_installations()
            if not any(d.samefile(k) for d in known_intallations)
        ]
        if available_installations:
            selection: tuple[Path, Flavour | None] | tuple[()] = ask(
                select(
                    'Installation:',
                    choices=[
                        *(
                            Choice(title=f'{k}  [{v}]', value=(k, v))
                            for k, v in available_installations
                        ),
                        skip,
                    ],
                )
            )
            if selection:
                (installation_path, flavour) = selection

                addon_dir = installation_path.joinpath(*ADDON_DIR_PARTS)
                addon_dir.mkdir(exist_ok=True)

                editable_config_values |= {
                    _EditableConfigOptions.AddonDir: addon_dir,
                    _EditableConfigOptions.GameFlavour: flavour,
                }
                installation_configured = 2 if flavour else 1

    if (
        installation_configured == 0
        and _EditableConfigOptions.AddonDir in interactive_editable_config_keys
    ):
        editable_config_values[_EditableConfigOptions.AddonDir] = ask(
            path(
                'Add-on directory:',
                only_directories=True,
                validate=AttrsFieldValidator(
                    fields(resolve_types(Config)).addon_dir,
                    config_converter,
                ),
            )
        )

    if (
        installation_configured < 2
        and _EditableConfigOptions.GameFlavour in interactive_editable_config_keys
    ):
        editable_config_values[_EditableConfigOptions.GameFlavour] = ask(
            select(
                'Game flavour:',
                choices=list(Flavour),
                initial_choice=config_values.get('addon_dir')
                and infer_flavour_from_addon_dir(config_values['addon_dir']),
            )
        )

    if _EditableConfigOptions.AutoUpdateCheck in interactive_editable_config_keys:
        editable_config_values[_EditableConfigOptions.AutoUpdateCheck] = ask(
            confirm('Periodically check for instawow updates?')
        )

    if _EditableConfigOptions.GithubAccessToken in interactive_editable_config_keys and ask(
        confirm('Set up GitHub authentication?')
    ):
        editable_config_values[_EditableConfigOptions.GithubAccessToken] = asyncio.run(
            asyncio.wait_for(_github_oauth_flow(), timeout=60 * 5)
        )

    if _EditableConfigOptions.CfcoreAccessToken in interactive_editable_config_keys:
        click.echo(
            textwrap.fill(
                'An access token is required to use CurseForge. '
                'Log in to https://console.curseforge.com/ to generate an access token.'
            )
        )
        editable_config_values[_EditableConfigOptions.CfcoreAccessToken] = (
            ask(password('CFCore access token:')) or None
        )

    if _EditableConfigOptions.WagoAddonsAccessToken in interactive_editable_config_keys:
        click.echo(
            textwrap.fill(
                'An access token is required to use Wago Addons. '
                'Wago issues tokens to Patreon subscribers above a certain tier. '
                'See https://addons.wago.io/patreon for more information.'
            )
        )
        editable_config_values[_EditableConfigOptions.WagoAddonsAccessToken] = (
            ask(password('Wago Addons access token:')) or None
        )

    for key, value in editable_config_values.items():
        parent = config_values
        for part in key.path[:-1]:
            parent = parent[part]
        parent[key.path[-1]] = value

    config = Config.from_values(config_values)
    config.global_config.write()
    config.write()

    click.echo('Configuration written to:')
    click.echo(f'  {config.global_config.config_file}')
    click.echo(f'  {config.config_file}')

    return config


@cli.group('weakauras-companion')
def _weakauras_group() -> None:
    "Manage your WeakAuras."


@_weakauras_group.command('build')
@click.pass_obj
def build_weakauras_companion(mw: _CtxObjWrapper) -> None:
    "Build the WeakAuras Companion add-on."
    from .wa_updater import WaCompanionBuilder

    mw.run_with_progress(WaCompanionBuilder(mw.manager.ctx).build())


@_weakauras_group.command('list')
@click.pass_obj
def list_installed_wago_auras(mw: _CtxObjWrapper) -> None:
    "List WeakAuras installed from Wago."
    from .wa_updater import WaCompanionBuilder

    aura_groups = WaCompanionBuilder(mw.manager.ctx).extract_installed_auras()
    installed_auras = sorted(
        (g.addon_name, a.id, a.url)
        for g in aura_groups
        for v in g.root.values()
        for a in v
        if not a.parent
    )
    click.echo(tabulate([('type', 'name', 'URL'), *installed_auras]))


@cli.command(hidden=True)
@click.option(
    '--start-date',
    callback=_parse_iso_date_into_datetime,
    help='Omit results before this date.',
    metavar='YYYY-MM-DD',
)
def generate_catalogue(start_date: datetime | None) -> None:
    "Generate the master catalogue."
    import json

    from .catalogue.cataloguer import Catalogue, catalogue_converter

    catalogue = asyncio.run(
        Catalogue.collate((r.catalogue for r in manager_ctx.ManagerCtx.RESOLVERS), start_date)
    )
    catalogue_json = catalogue_converter.unstructure(catalogue)

    catalogue_path = Path(f'base-catalogue-v{catalogue.version}.json').resolve()
    catalogue_path.write_text(
        json.dumps(catalogue_json, indent=2),
        encoding='utf-8',
    )
    catalogue_path.with_suffix(f'.compact{catalogue_path.suffix}').write_text(
        json.dumps(catalogue_json, separators=(',', ':')),
        encoding='utf-8',
    )


@cli.command
@click.pass_context
def gui(ctx: click.Context) -> None:
    "Fire up the GUI."
    from instawow_gui.app import InstawowApp

    global_config = GlobalConfig.read().ensure_dirs()
    dummy_jsonrpc_config = Config.make_dummy_config(
        global_config=global_config, profile='__jsonrpc__'
    ).ensure_dirs()

    params = ctx.find_root().params
    setup_logging(dummy_jsonrpc_config.logging_dir, *params['debug'])

    InstawowApp(
        debug=any(params['debug']), version=__version__
    ).main_loop()  # pyright: ignore[reportUnknownMemberType]



from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator, Callable, Generator, Sequence
from datetime import timedelta
from enum import IntEnum
from functools import partial
from itertools import count
from typing import Generic, TypeVar

import iso8601
from loguru import logger
from typing_extensions import NotRequired as N
from typing_extensions import TypedDict
from yarl import URL

from .. import pkg_models
from .. import results as R
from ..catalogue.cataloguer import CatalogueEntry
from ..common import ChangelogFormat, Defn, Flavour, SourceMetadata, Strategy
from ..config import GlobalConfig
from ..http import CACHE_INDEFINITELY, ClientSessionType, make_generic_progress_ctx
from ..resolvers import BaseResolver, HeadersIntent
from ..utils import gather, uniq

_T = TypeVar('_T')


_CF_WOW_GAME_ID = 1


class _CfCoreModLinks(TypedDict):
    websiteUrl: str
    wikiUrl: str
    issuesUrl: str
    sourceUrl: str


class _CfCoreModStatus(IntEnum):
    new = 1
    changed_required = 2
    under_soft_review = 3
    approved = 4
    rejected = 5
    changes_made = 6
    inactive = 7
    abandoned = 8
    deleted = 9
    under_review = 10


class _CfCoreCategory(TypedDict):
    id: int
    gameId: int
    name: str
    slug: str
    url: str
    iconUrl: str
    dateModified: str  # date-time
    isClass: bool
    classId: int
    parentCategoryId: int


class _CfCoreModAuthor(TypedDict):
    id: int
    name: str
    url: str


class _CfCoreModAsset(TypedDict):
    id: int
    modId: int
    title: str
    description: str
    thumbnailUrl: str
    url: str


class _CfCoreFileIndex(TypedDict):
    gameVersion: str
    fileId: int
    filename: str
    releaseType: _CfCoreFileReleaseType
    gameVersionTypeId: int
    modLoader: int


class _CfCoreFileStatus(IntEnum):
    processing = 1
    changes_required = 2
    under_review = 3
    approved = 4
    rejected = 5
    malware_detected = 6
    deleted = 7
    archived = 8
    testing = 9
    released = 10
    ready_for_review = 11
    deprecated = 12
    baking = 13
    awaiting_publishing = 14
    failed_publishing = 15


class _CfCoreFileReleaseType(IntEnum):
    release = 1
    beta = 2
    alpha = 3


class _CfCoreHashAlgo(IntEnum):
    sha1 = 1
    md5 = 2


class _CfCoreFileHash(TypedDict):
    value: str
    algo: _CfCoreHashAlgo


class _CfCoreSortableGameVersionTypeId(IntEnum):
    "Extracted from https://api.curseforge.com/v1/games/1/version-types."

    Retail = 517
    VanillaClassic = 67408
    Classic = 73713


class _CfCoreSortableGameVersion(TypedDict):
    gameVersionName: str
    gameVersionPadded: str
    gameVersion: str
    gameVersionReleaseDate: str  # date-time
    gameVersionTypeId: _CfCoreSortableGameVersionTypeId


class _CfCoreFileRelationType(IntEnum):
    embedded_library = 1
    optional_dependency = 2
    required_dependency = 3
    tool = 4
    incompatible = 5
    include = 6


class _CfCoreFileDependency(TypedDict):
    modId: int
    fileId: int
    relationType: _CfCoreFileRelationType


class _CfCoreFileModule(TypedDict):
    name: str
    fingerprint: int


class _CfCoreFile(TypedDict):
    id: int
    gameId: int
    modId: int
    isAvailable: bool
    displayName: str
    fileName: str
    releaseType: _CfCoreFileReleaseType
    fileStatus: _CfCoreFileStatus
    hashes: list[_CfCoreFileHash]
    fileDate: str  # date-time
    fileLength: int
    downloadCount: int
    downloadUrl: str | None  # null if distribution is forbidden
    gameVersions: list[str]
    sortableGameVersions: list[_CfCoreSortableGameVersion]
    dependencies: list[_CfCoreFileDependency]
    exposeAsAlternative: N[bool]
    parentProjectFileId: N[int]
    alternateFileId: int
    isServerPack: bool
    serverPackFileId: N[int]
    fileFingerprint: int
    modules: list[_CfCoreFileModule]


class _CfCoreMod(TypedDict):
    id: int
    gameId: int
    name: str
    slug: str
    links: _CfCoreModLinks
    summary: str
    status: _CfCoreModStatus
    downloadCount: int
    isFeatured: bool
    primaryCategoryId: int
    categories: list[_CfCoreCategory]
    authors: list[_CfCoreModAuthor]
    logo: _CfCoreModAsset
    screenshots: list[_CfCoreModAsset]
    mainFileId: int
    latestFiles: list[_CfCoreFile]
    latestFilesIndexes: list[_CfCoreFileIndex]
    dateCreated: str  # date-time
    dateModified: str  # date-time
    dateReleased: str  # date-time
    allowModDistribution: bool | None


class _CfCoreModsSearchSortField(IntEnum):
    featured = 1
    popularity = 2
    last_updated = 3
    name_ = 4
    author = 5
    total_downloads = 6
    category = 7
    game_version = 8


class _CfCoreResponsePagination(TypedDict):
    index: int
    pageSize: int
    resultCount: int
    totalCount: int | None


class _CfCorePaginatedResponse(TypedDict, Generic[_T]):
    data: _T
    pagination: _CfCoreResponsePagination


class _CfCoreStringDataResponse(TypedDict):
    data: str


class _CfCoreUnpaginatedModsResponse(TypedDict):
    data: list[_CfCoreMod]


class _CfCoreModsResponse(_CfCorePaginatedResponse['list[_CfCoreMod]']):
    pass


class _CfCoreFilesResponse(_CfCorePaginatedResponse['list[_CfCoreFile]']):
    pass


class CfCoreResolver(BaseResolver):
    metadata = SourceMetadata(
        id='curse',
        name='CFCore',
        strategies=frozenset(
            {
                Strategy.AnyReleaseType,
                Strategy.AnyFlavour,
                Strategy.VersionEq,
            }
        ),
        changelog_format=ChangelogFormat.Html,
        addon_toc_key='X-Curse-Project-ID',
    )
    requires_access_token = 'cfcore'

    # Ref: https://docs.curseforge.com/
    _mod_api_url = URL('https://api.curseforge.com/v1/mods')

    @classmethod
    def get_alias_from_url(cls, url: URL) -> str | None:
        if (
            url.host == 'www.curseforge.com'
            and len(url.parts) > 3
            and url.parts[1:3] == ('wow', 'addons')
        ):
            return url.parts[3].lower()

    async def make_request_headers(
        self, intent: HeadersIntent | None = None
    ) -> dict[str, str] | None:
        if intent is HeadersIntent.Download:
            return None

        maybe_access_token = self._get_access_token(self._manager_ctx.config.global_config)
        if maybe_access_token is None:
            raise ValueError(f'{self.metadata.name} access token is not configured')
        return {'x-api-key': maybe_access_token}

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, pkg_models.Pkg | R.ManagerError | R.InternalError]:
        numeric_ids = uniq(i for d in defns if (i := d.id or d.alias).isdigit())
        if not numeric_ids:
            return await super().resolve(defns)  # Fast path.

        async with self._manager_ctx.web_client.post(
            self._mod_api_url,
            expire_after=timedelta(minutes=5),
            headers=await self.make_request_headers(),
            json={'modIds': numeric_ids},
            raise_for_status=True,
        ) as response:
            response_json: _CfCoreUnpaginatedModsResponse = await response.json()

        addons_by_id = {str(r['id']): r for r in response_json['data']}

        results = await gather(
            (self.resolve_one(d, addons_by_id.get(d.id or d.alias)) for d in defns),
            R.resultify_async_exc,
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: _CfCoreMod | None) -> pkg_models.Pkg:
        if metadata is None:
            if defn.alias.isdigit():
                async with self._manager_ctx.web_client.get(
                    self._mod_api_url / defn.alias,
                    expire_after=timedelta(minutes=15),
                    headers=await self.make_request_headers(),
                ) as mod_response:
                    if mod_response.status == 404:
                        raise R.PkgNonexistent

                    mod_response.raise_for_status()

                    metadata = await mod_response.json()
                    assert metadata

            else:
                async with self._manager_ctx.web_client.get(
                    (self._mod_api_url / 'search').with_query(
                        gameId=_CF_WOW_GAME_ID, slug=defn.alias
                    ),
                    expire_after=timedelta(minutes=15),
                    headers=await self.make_request_headers(),
                    raise_for_status=True,
                ) as mod_response:
                    mod_response_json: _CfCoreModsResponse = await mod_response.json()
                    if not mod_response_json['data']:
                        raise R.PkgNonexistent

                    [metadata] = mod_response_json['data']

        if defn.strategies.version_eq:
            game_version_type_id = self._manager_ctx.config.game_flavour.to_flavour_keyed_enum(
                _CfCoreSortableGameVersionTypeId
            )
            files_url = (self._mod_api_url / str(metadata['id']) / 'files').with_query(
                gameVersionTypeId=game_version_type_id, pageSize=999
            )
            async with self._manager_ctx.web_client.get(
                files_url,
                expire_after=timedelta(hours=1),
                headers=await self.make_request_headers(),
                raise_for_status=True,
                trace_request_ctx=make_generic_progress_ctx(
                    f'Fetching metadata from {self.metadata.name}'
                ),
            ) as files_response:
                files_response_json: _CfCoreFilesResponse = await files_response.json()

            files = files_response_json['data']

        else:
            files = metadata['latestFiles']

        if not files:
            raise R.PkgFilesMissing

        def make_filter_fns() -> Generator[Callable[[_CfCoreFile], bool], None, None]:
            yield lambda f: not f.get('exposeAsAlternative', False)

            if not defn.strategies.any_flavour:
                type_id = self._manager_ctx.config.game_flavour.to_flavour_keyed_enum(
                    _CfCoreSortableGameVersionTypeId
                )
                yield lambda f: any(
                    s['gameVersionTypeId'] == type_id for s in f['sortableGameVersions']
                )

            if not defn.strategies.any_release_type:
                yield lambda f: f['releaseType'] == _CfCoreFileReleaseType.release

            if defn.strategies.version_eq:
                yield lambda f: f['displayName'] == defn.strategies.version_eq

        filter_fns = list(make_filter_fns())

        file = max(
            (f for f in files if all(r(f) for r in filter_fns)),
            # The ``id`` is just a counter so we don't have to go digging around dates
            key=lambda f: f['id'],
            default=None,
        )
        if file is None:
            raise R.PkgFilesNotMatching(defn.strategies)

        if file['downloadUrl'] is None:
            if metadata['allowModDistribution'] is False:
                raise R.PkgFilesMissing('package distribution is forbidden')
            else:
                raise R.PkgFilesMissing

        return pkg_models.Pkg(
            source=self.metadata.id,
            id=str(metadata['id']),
            slug=metadata['slug'],
            name=metadata['name'],
            description=metadata['summary'],
            url=metadata['links']['websiteUrl'],
            download_url=file['downloadUrl'],
            date_published=iso8601.parse_date(file['fileDate']),
            version=file['displayName'],
            changelog_url=str(
                self._mod_api_url / f'{metadata["id"]}/files/{file["id"]}/changelog'
            ),
            options=pkg_models.PkgOptions.from_strategy_values(defn.strategies),
            deps=[
                pkg_models.PkgDep(id=str(d['modId']))
                for d in file['dependencies']
                if d['relationType'] == _CfCoreFileRelationType.required_dependency
            ],
        )

    async def get_changelog(self, uri: URL) -> str:
        async with self._manager_ctx.web_client.get(
            uri,
            expire_after=CACHE_INDEFINITELY,
            headers=await self.make_request_headers(),
            raise_for_status=True,
        ) as response:
            response_json: _CfCoreStringDataResponse = await response.json()
            return response_json['data']

    @classmethod
    async def catalogue(cls, web_client: ClientSessionType) -> AsyncIterator[CatalogueEntry]:
        from aiohttp import ClientTimeout

        flavours_and_version_types = [
            (f, f.to_flavour_keyed_enum(_CfCoreSortableGameVersionTypeId)) for f in Flavour
        ]

        def excise_flavours(files: list[_CfCoreFile]):
            for flavour, version_type in flavours_and_version_types:
                if any(
                    not f.get('exposeAsAlternative', False)
                    and any(
                        s['gameVersionTypeId'] == version_type for s in f['sortableGameVersions']
                    )
                    for f in files
                ):
                    yield flavour

        def excise_folders(files: list[_CfCoreFile]):
            return uniq(frozenset(m['name'] for m in f['modules']) for f in files)

        STEP = 50
        MAX_OFFSET = 10_000  # The CF API craps out after 9,999 results

        get = partial(
            web_client.get,
            headers={'x-api-key': cls._get_access_token(GlobalConfig.from_values(env=True))},
            raise_for_status=True,
            timeout=ClientTimeout(total=10),
        )

        for offset in range(0, MAX_OFFSET, STEP):
            url = (cls._mod_api_url / 'search').with_query(
                gameId=_CF_WOW_GAME_ID,
                sortField=_CfCoreModsSearchSortField.last_updated,
                sortOrder='desc',
                pageSize=STEP,
                index=offset,
            )
            logger.debug(f'retrieving {url}')

            for attempt in count(1):
                try:
                    async with get(url) as response:
                        response_json: _CfCoreModsResponse = await response.json()
                        break
                except asyncio.TimeoutError:
                    logger.debug(f'request timed out; attempt {attempt} of 3')
                    if attempt < 3:
                        continue
            else:
                raise RuntimeError('maximum number of attempts exceeded')

            items = response_json['data']
            if not items:
                break

            for item in items:
                yield CatalogueEntry(
                    source=cls.metadata.id,
                    id=str(item['id']),
                    slug=item['slug'],
                    name=item['name'],
                    url=item['links']['websiteUrl'],
                    game_flavours=frozenset(excise_flavours(item['latestFiles'])),
                    download_count=item['downloadCount'],
                    last_updated=iso8601.parse_date(item['dateReleased']),
                    folders=excise_folders(item['latestFiles']),
                )
